# -*- coding: utf-8 -*-
#
#    fluxwallet - Python Cryptocurrency Library
#    Client for Bcoin Node
#    © 2019 June - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from fluxwallet.main import *
from fluxwallet.services.baseclient import BaseClient, ClientError
from fluxwallet.transactions import BitcoinTransaction, transaction_update_spents

PROVIDERNAME = "bcoin"

_logger = logging.getLogger(__name__)


class BcoinClient(BaseClient):
    """
    Class to interact with Bcoin API
    """

    def __init__(self, network, base_url, denominator, *args):
        super(self.__class__, self).__init__(
            network, PROVIDERNAME, base_url, denominator, *args
        )

    def compose_request(
        self, func, data="", parameter="", variables=None, method="get"
    ):
        url_path = func
        if data:
            url_path += "/" + str(data)
        if parameter:
            url_path += "/" + parameter
        if variables is None:
            variables = {}
        return self.request(url_path, variables, method, secure=False)

    def _parse_transaction(self, tx):
        status = "unconfirmed"
        if tx["confirmations"]:
            status = "confirmed"
        t = BitcoinTransaction.parse_hex(tx["hex"], strict=False, network=self.network)
        t.locktime = tx["locktime"]
        t.network = self.network
        t.fee = tx["fee"]
        t.date = datetime.utcfromtimestamp(tx["time"]) if tx["time"] else None
        t.confirmations = tx["confirmations"]
        t.block_height = tx["height"] if tx["height"] > 0 else None
        t.block_hash = tx["block"]
        t.status = status
        if not t.coinbase:
            for i in t.inputs:
                i.value = tx["inputs"][t.inputs.index(i)]["coin"]["value"]
        for o in t.outputs:
            o.spent = None
        t.update_totals()
        return t

    def getbalance(self, addresslist):
        balance = 0.0
        from fluxwallet.services.services import Service

        for address in addresslist:
            # First get all transactions for this address from the blockchain
            srv = Service(network=self.network.name, providers=["bcoin"])
            txs = srv.gettransactions(address, limit=25)

            # Fail if large number of transactions are found
            if not srv.complete:
                raise ClientError(
                    "If not all transactions are known, we cannot determine utxo's. "
                    "Increase limit or use other provider."
                )

            for a in [
                output for outputs in [t.outputs for t in txs] for output in outputs
            ]:
                if a.address == address:
                    balance += a.value
            for a in [i for inputs in [t.inputs for t in txs] for i in inputs]:
                if a.address == address:
                    balance -= a.value
        return int(balance)

    def getutxos(self, address, after_txid="", limit=MAX_TRANSACTIONS):
        # First get all transactions for this address from the blockchain
        txs = self.gettransactions(address, limit=50)

        # Fail if large number of transactions are found
        if len(txs) > 50:
            raise ClientError(
                "If not all transactions are known, we cannot determine utxo's"
            )

        utxos = []
        for tx in txs:
            for unspent in tx.outputs:
                if unspent.address != address:
                    continue
                if not self.isspent(tx.txid, unspent.output_n):
                    utxos.append(
                        {
                            "address": unspent.address,
                            "txid": tx.txid,
                            "confirmations": tx.confirmations,
                            "output_n": unspent.output_n,
                            "input_n": 0,
                            "block_height": tx.block_height,
                            "fee": tx.fee,
                            "size": tx.size,
                            "value": unspent.value,
                            "script": unspent.lock_script.hex(),
                            "date": tx.date,
                        }
                    )
                    if tx.txid == after_txid:
                        utxos = []
        return utxos[:limit]

    def gettransaction(self, txid):
        tx = self.compose_request("tx", txid)
        return self._parse_transaction(tx)

    def gettransactions(self, address, after_txid="", limit=MAX_TRANSACTIONS):
        assert limit > 0
        txs = []
        while True:
            variables = {"limit": limit, "after": after_txid}
            res = self.compose_request("tx", "address", address, variables)
            for tx in res:
                txs.append(self._parse_transaction(tx))
            if not txs or len(txs) >= limit:
                break
            if len(res) == limit:
                after_txid = res[limit - 1]["hash"]
            else:
                break

        # Check which outputs are spent/unspent for this address
        if not after_txid and len(txs) != limit:
            txs = transaction_update_spents(txs, address)
        return txs

    def getrawtransaction(self, txid):
        return self.compose_request("tx", txid)["hex"]

    def sendrawtransaction(self, rawtx):
        res = self.compose_request("broadcast", variables={"tx": rawtx}, method="post")
        txid = ""
        if "success" in res and res["success"]:
            t = BitcoinTransaction.parse_hex(rawtx, network=self.network)
            txid = t.txid
        return {"txid": txid, "response_dict": res}

    def estimatefee(self, blocks):
        if blocks > 15:
            blocks = 15
        fee = self.compose_request("fee", variables={"blocks": blocks})["rate"]
        if not fee:
            return False
        return fee

    def blockcount(self):
        return self.compose_request("")["chain"]["height"]

    def mempool(self, txid=""):
        txids = self.compose_request("mempool")
        if not txid:
            return txids
        elif txid in txids:
            return [txid]
        return []

    def getblock(self, blockid, parse_transactions, page, limit):
        block = self.compose_request("block", str(blockid))
        block["tx_count"] = len(block["txs"])
        txs = block["txs"]
        parsed_txs = []
        if parse_transactions:
            txs = txs[(page - 1) * limit : page * limit]
        for tx in txs:
            tx["confirmations"] = block["depth"]
            tx["time"] = block["time"]
            tx["height"] = block["height"]
            tx["block"] = block["hash"]
            if parse_transactions:
                t = self._parse_transaction(tx)
                parsed_txs.append(t)
            else:
                parsed_txs.append(tx["hash"])

        block["time"] = block["time"]
        block["txs"] = parsed_txs
        block["page"] = page
        block["pages"] = (
            None
            if not limit
            else int(block["tx_count"] // limit) + (block["tx_count"] % limit > 0)
        )
        block["limit"] = limit
        block["prev_block"] = block.pop("prevBlock")
        block["merkle_root"] = block.pop("merkleRoot")
        block["block_hash"] = block.pop("hash")
        return block

    def getrawblock(self, blockid):
        try:
            blockid = int(blockid)
            data = {"method": "getblockbyheight", "params": [blockid, 0, 0]}
        except ValueError:
            data = {"method": "getblock", "params": [blockid, 0, 0]}
        res = self.compose_request("", method="post", variables=data)
        return res["result"]

    def isspent(self, txid, index):
        try:
            self.compose_request("coin", txid, str(index))
        except ClientError:
            return 1
        return 0

    def getinfo(self):
        res = self.compose_request(
            "", variables={"method": "getmininginfo"}, method="post"
        )
        info = res["result"]
        return {
            "blockcount": info["blocks"],
            "chain": info["chain"],
            "difficulty": info["difficulty"],
            "hashrate": info["networkhashps"],
            "mempool_size": info["pooledtx"],
        }



import logging

_logger = logging.getLogger(__name__)


class WalletError(Exception):
    """
    Handle Wallet class Exceptions

    """

    def __init__(self, msg=""):
        self.msg = msg
        _logger.error(msg)

    def __str__(self):
        return self.msg



# -*- coding: utf-8 -*-
#
#    fluxwallet - Python Cryptocurrency Library
#    Chain.so client
#    © 2017-2022 October - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
from datetime import datetime

from fluxwallet.main import MAX_TRANSACTIONS
from fluxwallet.services.baseclient import BaseClient, ClientError
from fluxwallet.transactions import BitcoinTransaction

_logger = logging.getLogger(__name__)

PROVIDERNAME = "chainso"


class ChainSo(BaseClient):
    def __init__(self, network, base_url, denominator, *args):
        super(self.__class__, self).__init__(
            network, PROVIDERNAME, base_url, denominator, *args
        )

    def compose_request(
        self, function, data="", parameter="", variables=None, method="get"
    ):
        url_path = function
        url_path += "/" + self.provider_coin_id
        if data:
            url_path += "/" + data
        if parameter:
            url_path += "/" + parameter
        if variables is None:
            variables = {}
        if self.api_key:
            variables.update({"api_key": self.api_key})
        return self.request(url_path, variables, method)

    def sendrawtransaction(self, rawtx):
        res = self.compose_request(
            "send_tx", variables={"tx_hex": rawtx}, method="post"
        )
        return {
            "txid": "" if "data" not in res else res["data"]["txid"],
            "response_dict": res,
        }

    def getbalance(self, addresslist):
        balance = 0.0
        for address in addresslist:
            res = self.compose_request("get_address_balance", address)
            balance += float(res["data"]["confirmed_balance"]) + float(
                res["data"]["unconfirmed_balance"]
            )
        return int(balance * self.units)

    def getutxos(self, address, after_txid="", limit=MAX_TRANSACTIONS):
        txs = []
        lasttx = after_txid
        res = self.compose_request("get_tx_unspent", address, lasttx)
        if res["status"] != "success":
            pass
        for tx in res["data"]["txs"][:limit]:
            txs.append(
                {
                    "address": address,
                    "txid": tx["txid"],
                    "confirmations": tx["confirmations"],
                    "output_n": -1 if "output_no" not in tx else tx["output_no"],
                    "input_n": -1 if "input_no" not in tx else tx["input_no"],
                    "block_height": None,
                    "fee": None,
                    "size": 0,
                    "value": int(round(float(tx["value"]) * self.units, 0)),
                    "script": tx["script_hex"],
                    "date": datetime.utcfromtimestamp(tx["time"]),
                }
            )
        if len(txs) >= 1000:
            _logger.warning(
                "ChainSo: transaction list has been truncated, and thus is incomplete"
            )
        return txs

    def getrawtransaction(self, txid):
        res = self.compose_request("get_tx", txid)
        return res["data"]["tx_hex"]

    def gettransaction(self, txid, block_height=None):
        res = self.compose_request("get_tx", txid)
        tx = res["data"]
        rawtx = tx["tx_hex"]
        t = BitcoinTransaction.parse_hex(
            rawtx, strict=self.strict, network=self.network
        )
        input_total = 0
        output_total = 0
        if not t.coinbase:
            for n, i in enumerate(t.inputs):
                i.value = int(round(float(tx["inputs"][n]["value"]) * self.units, 0))
                input_total += i.value
        for o in t.outputs:
            o.spent = None
            output_total += o.value
        if not t.block_height and tx["confirmations"]:
            t.block_height = self.getblock(tx["blockhash"], False, 1, 1)["height"]
        t.block_hash = tx["blockhash"]
        t.rawtx = bytes.fromhex(rawtx)
        t.size = tx["size"]
        t.network = self.network
        t.locktime = tx["locktime"]
        t.input_total = input_total
        t.output_total = output_total
        t.fee = 0
        if t.input_total:
            t.fee = t.input_total - t.output_total
        t.confirmations = tx["confirmations"]
        if tx["confirmations"]:
            t.status = "confirmed"
            t.date = datetime.utcfromtimestamp(tx["time"])
        else:
            t.status = "unconfirmed"
            t.date = None
        return t

    def gettransactions(self, address, after_txid="", limit=MAX_TRANSACTIONS):
        txs = []
        res1 = self.compose_request("get_tx_received", address, after_txid)
        if res1["status"] != "success":
            raise ClientError(
                "Chainso get_tx_received request unsuccessful, status: %s"
                % res1["status"]
            )
        res2 = self.compose_request("get_tx_spent", address, after_txid)
        if res2["status"] != "success":
            raise ClientError(
                "Chainso get_tx_spent request unsuccessful, status: %s" % res2["status"]
            )
        res = res1["data"]["txs"] + res2["data"]["txs"]
        res = sorted(res, key=lambda x: x["time"])
        tx_conf = []
        for t in res:
            tt = (t["confirmations"], t["txid"])
            if tt not in tx_conf:
                tx_conf.append(tt)
        for tx in tx_conf[:limit]:
            t = self.gettransaction(tx[1])
            txs.append(t)
        return txs

    def blockcount(self):
        return self.compose_request("get_info")["data"]["blocks"]

    def mempool(self, txid):
        res = self.compose_request("is_tx_confirmed", txid)
        if res["status"] == "success" and res["data"]["confirmations"] == 0:
            return [txid]
        return []

    def getblock(self, blockid, parse_transactions, page, limit):
        if limit > 5:
            limit = 5
        bd = self.compose_request("get_block", str(blockid))["data"]
        if parse_transactions:
            txs = []
            for txid in bd["txs"][(page - 1) * limit : page * limit]:
                # try:
                txs.append(self.gettransaction(txid, block_height=bd["block_no"]))
                # except Exception as e:
                #     raise ClientError("Could not parse tx %s with error %s" % (txid, e))
        else:
            txs = bd["txs"]

        n_txs = len(bd["txs"])
        block = {
            "bits": None,
            "depth": bd["confirmations"],
            "block_hash": bd["blockhash"],
            "height": bd["block_no"],
            "merkle_root": bd["merkleroot"],
            "nonce": None,
            "prev_block": bd["previous_blockhash"],
            "time": bd["time"],
            "tx_count": n_txs,
            "txs": txs,
            "version": b"",
            "page": page,
            "pages": None if not limit else int(n_txs // limit) + (n_txs % limit > 0),
            "limit": limit,
        }
        return block

    # def getrawblock(self, blockid):

    # def isspent(self, txid, output_n):

    def getinfo(self):
        info = self.compose_request("get_info")["data"]
        return {
            "blockcount": info["blocks"],
            "chain": info["name"],
            "difficulty": int(float(info["mining_difficulty"])),
            "hashrate": int(float(info["hashrate"])),
            "mempool_size": int(info["unconfirmed_txs"]),
        }



# -*- coding: utf-8 -*-
#
#    fluxwallet - Python Cryptocurrency Library
#    DataBase - SqlAlchemy database definitions
#    © 2016 - 2022 October - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import annotations

from urllib.parse import urlparse

from sqlalchemy import (
    BigInteger,
    Boolean,
    CheckConstraint,
    Column,
    DateTime,
    ForeignKey,
    Integer,
    LargeBinary,
    Sequence,
    String,
    UniqueConstraint,
    create_engine,
)
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.orm import (
    Mapped,
    close_all_sessions,
    declarative_base,
    relationship,
    sessionmaker,
)

from fluxwallet.main import *

_logger = logging.getLogger(__name__)
Base = declarative_base()


@compiles(LargeBinary, "mysql")
def compile_largebinary_mysql(type_, compiler, **kwargs):
    length = type_.length
    element = "BLOB" if not length else "VARBINARY(%d)" % length
    return element


class Db:
    """
    fluxwallet Database object used by Service() and HDWallet() class. Initialize database and open session when
    creating database object.

    Create new database if is doesn't exist yet

    """

    def __enter__(self):
        return self.session

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.session.close()

    def __init__(self, db_uri=None, password=None):
        if db_uri is None:
            db_uri = DEFAULT_DATABASE
        self.o = urlparse(db_uri)
        if (
            not self.o.scheme or len(self.o.scheme) < 2
        ):  # Dirty hack to avoid issues with urlparse on Windows confusing drive with scheme
            if password:
                # Warning: This requires the pysqlcipher3 module
                db_uri = (
                    "sqlite+pysqlcipher://:%s@/%s?cipher=aes-256-cfb&kdf_iter=64000"
                    % (password, db_uri)
                )
            else:
                db_uri = "sqlite:///%s" % db_uri

        if db_uri.startswith("sqlite://") and ALLOW_DATABASE_THREADS:
            db_uri += "&" if "?" in db_uri else "?"
            db_uri += "check_same_thread=False"
        if self.o.scheme == "mysql":
            db_uri += "&" if "?" in db_uri else "?"
            db_uri += "binary_prefix=true"
        self.engine = create_engine(db_uri, isolation_level="READ UNCOMMITTED")

        Session = sessionmaker(bind=self.engine)
        Base.metadata.create_all(self.engine)
        self._import_config_data(Session)
        self.session = Session()

        _logger.info(
            "Using database: %s://%s:%s/%s"
            % (
                self.o.scheme or "",
                self.o.hostname or "",
                self.o.port or "",
                self.o.path or "",
            )
        )
        self.db_uri = db_uri

        # VERIFY AND UPDATE DATABASE
        # Just a very simple database update script, without any external libraries for now
        #
        version_db = (
            self.session.query(DbConfig.value).filter_by(variable="version").scalar()
        )
        if version_db[:3] == "0.4" and FLUXWALLET_VERSION[:3] >= "0.5":
            raise ValueError(
                "Old database version found (<0.4.19). Can not convert to >0.5 version database "
                "automatically, use updatedb.py tool to update"
            )
        try:
            if FLUXWALLET_VERSION != version_db:
                _logger.warning(
                    "fluxwallet database (%s) is from different version then library code (%s). "
                    "Let's try to update database." % (version_db, FLUXWALLET_VERSION)
                )
                db_update(self, version_db, FLUXWALLET_VERSION)
        except Exception as e:
            _logger.warning("Error when verifying version or updating database: %s" % e)

    def drop_db(self, yes_i_am_sure=False):
        if yes_i_am_sure:
            self.session.commit()
            self.session.close_all()
            close_all_sessions()
            Base.metadata.drop_all(self.engine)

    @staticmethod
    def _import_config_data(ses):
        session = ses()
        installation_date = (
            session.query(DbConfig.value)
            .filter_by(variable="installation_date")
            .scalar()
        )
        if not installation_date:
            session.merge(DbConfig(variable="version", value=FLUXWALLET_VERSION))
            session.merge(
                DbConfig(variable="installation_date", value=str(datetime.now()))
            )
            url = ""
            try:
                url = str(session.bind.url)
            except Exception:
                pass
            session.merge(DbConfig(variable="installation_url", value=url))
            session.commit()
        session.close()


def add_column(engine, table_name, column):
    """
    Used to add new column to database with migration and update scripts

    :param engine:
    :param table_name:
    :param column:
    :return:
    """
    column_name = column.compile(dialect=engine.dialect)
    column_type = column.type.compile(engine.dialect)
    engine.execute(
        "ALTER TABLE %s ADD COLUMN %s %s" % (table_name, column_name, column_type)
    )


class DbAddressBook(Base):
    """
    Fluxwallet address book mapping
    """

    __tablename__ = "address_book"
    id = Column(
        Integer,
        Sequence("addressbook_id_seq"),
        primary_key=True,
        doc="Unique addressbook ID",
    )
    name = Column(String(80), unique=True, doc="Unique address nickname")
    address = Column(
        String(255),
        index=True,
        doc="Address representation of key. A cryptocurrency address is a hash of the public key",
    )


class DbConfig(Base):
    """
    fluxwallet configuration variables

    """

    __tablename__ = "config"
    variable = Column(String(30), primary_key=True)
    value = Column(String(255))


class DbWallet(Base):
    """
    Database definitions for wallets in Sqlalchemy format

    Contains one or more keys.

    """

    __tablename__ = "wallets"
    id = Column(
        Integer, Sequence("wallet_id_seq"), primary_key=True, doc="Unique wallet ID"
    )
    name = Column(String(80), unique=True, doc="Unique wallet name")
    owner = Column(String(50), doc="Wallet owner")
    network_name = Column(
        String(20),
        ForeignKey("networks.name"),
        doc="Name of network, i.e.: bitcoin, litecoin",
    )
    network: Mapped[DbNetwork] = relationship(doc="Link to DbNetwork object")
    purpose = Column(
        Integer,
        doc="Wallet purpose ID. BIP-44 purpose field, indicating which key-scheme is used default is 44",
    )
    scheme = Column(String(25), doc="Key structure type, can be BIP-32 or single")
    witness_type = Column(
        String(20),
        default="legacy",
        doc="Wallet witness type. Can be 'legacy', 'segwit' or 'p2sh-segwit'. Default is legacy.",
    )
    encoding = Column(
        String(15),
        default="base58",
        doc="Default encoding to use for address generation, i.e. base58 or bech32. Default is base58.",
    )
    main_key_id = Column(
        Integer,
        doc="Masterkey ID for this wallet. All other keys are derived from the masterkey in a "
        "HD wallet bip32 wallet",
    )
    keys: Mapped[DbKey] = relationship(
        back_populates="wallet",
        doc="Link to keys (DbKeys objects) in this wallet",
    )
    transactions: Mapped[DbTransaction] = relationship(
        back_populates="wallet",
        doc="Link to transaction (DbTransactions) in this wallet",
    )
    multisig_n_required = Column(
        Integer,
        default=1,
        doc="Number of required signature for multisig, "
        "only used for multisignature master key",
    )
    sort_keys = Column(Boolean, default=False, doc="Sort keys in multisig wallet")
    parent_id = Column(
        Integer,
        ForeignKey("wallets.id"),
        doc="Wallet ID of parent wallet, used in multisig wallets",
    )
    children: Mapped[DbWallet] = relationship(
        # lazy="joined",
        # join_depth=2,
        doc="Wallet IDs of children wallets, used in multisig wallets",
    )
    multisig = Column(
        Boolean,
        default=True,
        doc="Indicates if wallet is a multisig wallet. Default is True",
    )
    cosigner_id = Column(
        Integer,
        doc="ID of cosigner of this wallet. Used in multisig wallets to differentiate between "
        "different wallets",
    )
    key_path = Column(
        String(100),
        doc="Key path structure used in this wallet. Key path for multisig wallet, use to create "
        "your own non-standard key path. Key path must follow the following rules: "
        "* Path start with masterkey (m) and end with change / address_index "
        "* If accounts are used, the account level must be 3. I.e.: m/purpose/coin_type/account/ "
        "* All keys must be hardened, except for change, address_index or cosigner_id "
        " Max length of path is 8 levels",
    )
    default_account_id = Column(
        Integer,
        doc="ID of default account for this wallet if multiple accounts are used",
    )

    __table_args__ = (
        CheckConstraint(
            scheme.in_(["single", "bip32"]), name="constraint_allowed_schemes"
        ),
        CheckConstraint(
            encoding.in_(["base58", "bech32"]),
            name="constraint_default_address_encodings_allowed",
        ),
        CheckConstraint(
            witness_type.in_(["legacy", "segwit", "p2sh-segwit"]),
            name="wallet_constraint_allowed_types",
        ),
    )

    def __repr__(self):
        return "<DbWallet(name='%s', network='%s'>" % (self.name, self.network_name)


class DbKeyMultisigChildren(Base):
    """
    Use many-to-many relationship for multisig keys. A multisig keys contains 2 or more child keys
    and a child key can be used in more then one multisig key.

    """

    __tablename__ = "key_multisig_children"

    parent_id = Column(Integer, ForeignKey("keys.id"), primary_key=True)
    child_id = Column(Integer, ForeignKey("keys.id"), primary_key=True)
    key_order = Column(Integer, Sequence("key_multisig_children_id_seq"))


class DbKey(Base):
    """
    Database definitions for keys in Sqlalchemy format

    Part of a wallet, and used by transactions

    """

    __tablename__ = "keys"
    id = Column(Integer, Sequence("key_id_seq"), primary_key=True, doc="Unique Key ID")
    parent_id = Column(
        Integer, Sequence("parent_id_seq"), doc="Parent Key ID. Used in HD wallets"
    )
    name = Column(String(80), index=True, doc="Key name string")
    account_id = Column(
        Integer, index=True, doc="ID of account if key is part of a HD structure"
    )
    depth = Column(
        Integer,
        doc="Depth of key if it is part of a HD structure. Depth=0 means masterkey, "
        "depth=1 are the masterkeys children.",
    )
    change = Column(Integer, doc="Change or normal address: Normal=0, Change=1")
    address_index = Column(
        BigInteger, doc="Index of address in HD key structure address level"
    )
    public = Column(
        LargeBinary(128), index=True, doc="Bytes representation of public key"
    )
    private = Column(
        LargeBinary(128), index=True, doc="Bytes representation of private key"
    )
    wif = Column(
        String(255),
        index=True,
        doc="Public or private WIF (Wallet Import Format) representation",
    )
    compressed = Column(
        Boolean, default=True, doc="Is key compressed or not. Default is True"
    )
    key_type = Column(
        String(10),
        default="bip32",
        doc="Type of key: single, bip32 or multisig. Default is bip32",
    )
    address = Column(
        String(255),
        index=True,
        doc="Address representation of key. An cryptocurrency address is a hash of the public key",
    )
    cosigner_id = Column(
        Integer, doc="ID of cosigner, used if key is part of HD Wallet"
    )
    encoding = Column(
        String(15),
        default="base58",
        doc="Encoding used to represent address: base58 or bech32",
    )
    purpose = Column(Integer, default=44, doc="Purpose ID, default is 44")
    is_private = Column(Boolean, doc="Is key private or not?")
    path = Column(String(100), doc="String of BIP-32 key path")
    wallet_id = Column(
        Integer,
        ForeignKey("wallets.id"),
        index=True,
        doc="Wallet ID which contains this key",
    )
    wallet = relationship(
        "DbWallet", back_populates="keys", doc="Related Wallet object"
    )
    transaction_inputs = relationship(
        "DbTransactionInput",
        cascade="all,delete",
        back_populates="key",
        doc="All DbTransactionInput objects this key is part of",
    )
    transaction_outputs = relationship(
        "DbTransactionOutput",
        cascade="all,delete",
        back_populates="key",
        doc="All DbTransactionOutput objects this key is part of",
    )
    balance = Column(
        BigInteger, default=0, doc="Total balance of UTXO's linked to this key"
    )
    used = Column(
        Boolean,
        default=False,
        doc="Has key already been used on the blockchain in as input or output? "
        "Default is False",
    )
    network_name = Column(
        String(20),
        ForeignKey("networks.name"),
        doc="Name of key network, i.e. bitcoin, litecoin, dash",
    )
    latest_txid = Column(
        LargeBinary(32), doc="TxId of latest transaction downloaded from the blockchain"
    )
    network = relationship("DbNetwork", doc="DbNetwork object for this key")
    multisig_parents = relationship(
        "DbKeyMultisigChildren",
        backref="child_key",
        primaryjoin=id == DbKeyMultisigChildren.child_id,
        doc="List of parent keys",
    )
    multisig_children = relationship(
        "DbKeyMultisigChildren",
        backref="parent_key",
        order_by="DbKeyMultisigChildren.key_order",
        primaryjoin=id == DbKeyMultisigChildren.parent_id,
        doc="List of children keys",
    )

    __table_args__ = (
        CheckConstraint(
            key_type.in_(["single", "bip32", "multisig"]),
            name="constraint_key_types_allowed",
        ),
        CheckConstraint(
            encoding.in_(["base58", "bech32"]),
            name="constraint_address_encodings_allowed",
        ),
        UniqueConstraint("wallet_id", "public", name="constraint_wallet_pubkey_unique"),
        UniqueConstraint(
            "wallet_id", "private", name="constraint_wallet_privkey_unique"
        ),
        UniqueConstraint("wallet_id", "wif", name="constraint_wallet_wif_unique"),
        UniqueConstraint(
            "wallet_id", "address", name="constraint_wallet_address_unique"
        ),
    )

    def __repr__(self):
        return "<DbKey(id='%s', name='%s', wif='%s'>" % (self.id, self.name, self.wif)


class DbNetwork(Base):
    """
    Database definitions for networks in Sqlalchemy format

    Most network settings and variables can be found outside the database in the libraries configurations settings.
    Use the fluxwallet/data/networks.json file to view and manage settings.

    """

    __tablename__ = "networks"
    name = Column(
        String(20),
        unique=True,
        primary_key=True,
        doc="Network name, i.e.: bitcoin, litecoin, dash",
    )
    description = Column(String(50))

    def __repr__(self):
        return "<DbNetwork(name='%s', description='%s'>" % (self.name, self.description)


# class TransactionType(enum.Enum):
#     """
#     Incoming or Outgoing transaction Enumeration
#     """
#     incoming = 1
#     outgoing = 2


class DbTransaction(Base):
    """
    Database definitions for transactions in Sqlalchemy format

    Refers to 1 or more keys which can be part of a wallet

    """

    __tablename__ = "transactions"
    id = Column(
        Integer,
        Sequence("transaction_id_seq"),
        primary_key=True,
        doc="Unique transaction index for internal usage",
    )
    txid = Column(
        LargeBinary(32), index=True, doc="Bytes representation of transaction ID"
    )
    wallet_id = Column(
        Integer,
        ForeignKey("wallets.id"),
        index=True,
        doc="ID of wallet which contains this transaction",
    )
    account_id = Column(Integer, index=True, doc="ID of account")
    wallet = relationship(
        "DbWallet",
        back_populates="transactions",
        doc="Link to Wallet object which contains this transaction",
    )
    witness_type = Column(
        String(20), default="legacy", doc="Is this a legacy or segwit transaction?"
    )
    version = Column(
        BigInteger,
        default=1,
        doc="Tranaction version. Default is 1 but some wallets use another version number",
    )
    locktime = Column(
        BigInteger,
        default=0,
        doc="Transaction level locktime. Locks the transaction until a specified block "
        "(value from 1 to 5 million) or until a certain time (Timestamp in seconds after 1-jan-1970)."
        " Default value is 0 for transactions without locktime",
    )
    date = Column(
        DateTime,
        default=datetime.utcnow,
        doc="Date when transaction was confirmed and included in a block. "
        "Or when it was created when transaction is not send or confirmed",
    )
    coinbase = Column(
        Boolean,
        default=False,
        doc="Is True when this is a coinbase transaction, default is False",
    )
    expiry_height = Column(
        Integer,
        default=0,
        doc="Expiry height (in blocktime) for this transactions. I.e, must be confirmed before this block height."
        "Default is 0: no expiry (not version 4 transaction)",
    )
    confirmations = Column(
        Integer,
        default=0,
        doc="Number of confirmation when this transaction is included in a block. "
        "Default is 0: unconfirmed",
    )
    block_height = Column(
        Integer, index=True, doc="Number of block this transaction is included in"
    )
    size = Column(Integer, doc="Size of the raw transaction in bytes")
    fee = Column(BigInteger, doc="Transaction fee")
    inputs = relationship(
        "DbTransactionInput",
        cascade="all,delete",
        doc="List of all inputs as DbTransactionInput objects",
    )
    outputs = relationship(
        "DbTransactionOutput",
        cascade="all,delete",
        doc="List of all outputs as DbTransactionOutput objects",
    )
    status = Column(
        String(20),
        default="new",
        doc="Current status of transaction, can be one of the following: new', "
        "'unconfirmed', 'confirmed'. Default is 'new'",
    )
    is_complete = Column(
        Boolean,
        default=True,
        doc="Allow to store incomplete transactions, for instance if not all "
        "inputs are known when retrieving UTXO's",
    )
    input_total = Column(
        BigInteger,
        default=0,
        doc="Total value of the inputs of this transaction. Input total = Output total + fee. "
        "Default is 0",
    )
    output_total = Column(
        BigInteger,
        default=0,
        doc="Total value of the outputs of this transaction. Output total = Input total - fee",
    )
    network_name = Column(
        String(20),
        ForeignKey("networks.name"),
        doc="Blockchain network name of this transaction",
    )
    network = relationship("DbNetwork", doc="Link to DbNetwork object")
    raw = Column(
        LargeBinary,
        doc="Raw transaction hexadecimal string. Transaction is included in raw format on the blockchain",
    )
    verified = Column(
        Boolean, default=False, doc="Is transaction verified. Default is False"
    )

    __table_args__ = (
        UniqueConstraint(
            "wallet_id", "txid", name="constraint_wallet_transaction_hash_unique"
        ),
        CheckConstraint(
            status.in_(["new", "unconfirmed", "confirmed"]),
            name="constraint_status_allowed",
        ),
        CheckConstraint(
            witness_type.in_(["legacy", "segwit"]),
            name="transaction_constraint_allowed_types",
        ),
    )

    def __repr__(self):
        return "<DbTransaction(txid='%s', confirmations='%s')>" % (
            self.txid,
            self.confirmations,
        )


class DbTransactionInput(Base):
    """
    Transaction Input Table

    Relates to Transaction table and Key table

    """

    __tablename__ = "transaction_inputs"
    transaction_id = Column(
        Integer,
        ForeignKey("transactions.id"),
        primary_key=True,
        doc="Input is part of transaction with this ID",
    )
    transaction = relationship(
        "DbTransaction", back_populates="inputs", doc="Related DbTransaction object"
    )
    index_n = Column(Integer, primary_key=True, doc="Index number of transaction input")
    key_id = Column(
        Integer, ForeignKey("keys.id"), index=True, doc="ID of key used in this input"
    )
    key = relationship(
        "DbKey", back_populates="transaction_inputs", doc="Related DbKey object"
    )
    address = Column(
        String(255),
        doc="Address string of input, used if no key is associated. "
        "An cryptocurrency address is a hash of the public key or a redeemscript",
    )
    witnesses = Column(
        LargeBinary, doc="Witnesses (signatures) used in Segwit transaction inputs"
    )
    witness_type = Column(
        String(20),
        default="legacy",
        doc="Type of transaction, can be legacy, segwit or p2sh-segwit. Default is legacy",
    )
    prev_txid = Column(
        LargeBinary(32),
        doc="Transaction hash of previous transaction. Previous unspent outputs (UTXO) is spent "
        "in this input",
    )
    output_n = Column(
        BigInteger,
        doc="Output_n of previous transaction output that is spent in this input",
    )
    script = Column(
        LargeBinary, doc="Unlocking script to unlock previous locked output"
    )
    script_type = Column(
        String(20),
        default="sig_pubkey",
        doc="Unlocking script type. Can be 'coinbase', 'sig_pubkey', 'p2sh_multisig', 'signature', "
        "'unknown', 'p2sh_p2wpkh' or 'p2sh_p2wsh'. Default is sig_pubkey",
    )
    sequence = Column(
        BigInteger,
        doc="Transaction sequence number. Used for timelock transaction inputs",
    )
    value = Column(BigInteger, default=0, doc="Value of transaction input")
    double_spend = Column(
        Boolean,
        default=False,
        doc="Indicates if a service provider tagged this transaction as double spend",
    )

    __table_args__ = (
        CheckConstraint(
            witness_type.in_(["legacy", "segwit", "p2sh-segwit"]),
            name="transactioninput_constraint_allowed_types",
        ),
        UniqueConstraint(
            "transaction_id", "index_n", name="constraint_transaction_input_unique"
        ),
    )


class DbTransactionOutput(Base):
    """
    Transaction Output Table

    Relates to Transaction and Key table

    When spent is False output is considered an UTXO

    """

    __tablename__ = "transaction_outputs"
    transaction_id = Column(
        Integer,
        ForeignKey("transactions.id"),
        primary_key=True,
        doc="Transaction ID of parent transaction",
    )
    transaction = relationship(
        "DbTransaction", back_populates="outputs", doc="Link to transaction object"
    )
    output_n = Column(
        Integer, primary_key=True, doc="Sequence number of transaction output"
    )
    key_id = Column(
        Integer,
        ForeignKey("keys.id"),
        index=True,
        doc="ID of key used in this transaction output",
    )
    key = relationship(
        "DbKey",
        back_populates="transaction_outputs",
        doc="List of DbKey object used in this output",
    )
    address = Column(
        String(255),
        doc="Address string of output, used if no key is associated. "
        "An cryptocurrency address is a hash of the public key or a redeemscript",
    )
    script = Column(LargeBinary, doc="Locking script which locks transaction output")
    script_type = Column(
        String(20),
        default="p2pkh",
        doc="Locking script type. Can be one of these values: 'p2pkh', 'multisig', 'p2sh', 'p2pk', "
        "'nulldata', 'unknown', 'p2wpkh', 'p2wsh', 'p2tr'. Default is p2pkh",
    )
    value = Column(BigInteger, default=0, doc="Total transaction output value")
    spent = Column(
        Boolean,
        default=False,
        doc="Indicated if output is already spent in another transaction",
    )
    spending_txid = Column(
        LargeBinary(32), doc="Transaction hash of input which spends this output"
    )
    spending_index_n = Column(
        Integer, doc="Index number of transaction input which spends this output"
    )

    __table_args__ = (
        UniqueConstraint(
            "transaction_id", "output_n", name="constraint_transaction_output_unique"
        ),
    )


def db_update_version_id(db, version):
    _logger.info("Updated fluxwallet database to version %s" % version)
    db.session.query(DbConfig).filter(DbConfig.variable == "version").update(
        {DbConfig.value: version}
    )
    db.session.commit()
    return version


def db_update(db, version_db, code_version=FLUXWALLET_VERSION):
    # Database changes from version 0.5+
    #
    if version_db <= "0.6.3" and code_version > "0.6.3":
        # Example: column = Column('latest_txid', String(32))
        column = Column(
            "witnesses",
            LargeBinary,
            doc="Witnesses (signatures) used in Segwit transaction inputs",
        )
        add_column(db.engine, "transaction_inputs", column)
        # version_db = db_update_version_id(db, '0.6.4')
    version_db = db_update_version_id(db, code_version)
    return version_db



# -*- coding: utf-8 -*-
#
#    fluxwallet - Python Cryptocurrency Library
#    MAIN - Load configs, initialize logging and database
#    © 2017 - 2022 October - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import functools
import logging

# Do not remove any of the imports below, used by other files
import os
import sys
from logging.handlers import RotatingFileHandler

from fluxwallet.config.config import *

# Initialize logging
logger = logging.getLogger("fluxwallet")
logger.setLevel(LOGLEVEL)

if ENABLE_fluxwallet_LOGGING:
    handler = RotatingFileHandler(
        str(FW_LOG_FILE), maxBytes=100 * 1024 * 1024, backupCount=2
    )
    formatter = logging.Formatter(
        "%(asctime)s %(levelname)s %(funcName)s(%(lineno)d) %(message)s",
        datefmt="%Y/%m/%d %H:%M:%S",
    )
    handler.setFormatter(formatter)
    handler.setLevel(LOGLEVEL)
    logger.addHandler(handler)

    _logger = logging.getLogger(__name__)
    logger.info("WELCOME TO fluxwallet - CRYPTOCURRENCY LIBRARY")
    logger.info("Version: %s" % FLUXWALLET_VERSION)
    logger.info("Logger name: %s" % logging.__name__)
    logger.info("Read config from: %s" % FW_CONFIG_FILE)
    logger.info("Directory databases: %s" % FW_DATABASE_DIR)
    logger.info("Default database: %s" % DEFAULT_DATABASE)
    logger.info("Logging to: %s" % FW_LOG_FILE)
    logger.info("Directory for data files: %s" % FW_DATA_DIR)


def script_type_default(witness_type=None, multisig=False, locking_script=False):
    """
    Determine default script type for provided witness type and key type combination used in this library.

    >>> script_type_default('segwit', locking_script=True)
    'p2wpkh'

    :param witness_type: Witness type used: standard, p2sh-segwit or segwit
    :type witness_type: str
    :param multisig: Multi-signature key or not, default is False
    :type multisig: bool
    :param locking_script: Limit search to locking_script. Specify False for locking scripts and True for unlocking scripts
    :type locking_script: bool

    :return str: Default script type
    """

    if not witness_type:
        return None
    if witness_type == "legacy" and not multisig:
        return "p2pkh" if locking_script else "sig_pubkey"
    elif witness_type == "legacy" and multisig:
        return "p2sh" if locking_script else "p2sh_multisig"
    elif witness_type == "segwit" and not multisig:
        return "p2wpkh" if locking_script else "sig_pubkey"
    elif witness_type == "segwit" and multisig:
        return "p2wsh" if locking_script else "p2sh_multisig"
    elif witness_type == "p2sh-segwit" and not multisig:
        return "p2sh" if locking_script else "p2sh_p2wpkh"
    elif witness_type == "p2sh-segwit" and multisig:
        return "p2sh" if locking_script else "p2sh_p2wsh"
    else:
        raise ValueError(
            "Wallet and key type combination not supported: %s / %s"
            % (witness_type, multisig)
        )


def get_encoding_from_witness(witness_type=None):
    """
    Derive address encoding (base58 or bech32) from transaction witness type.

    Returns 'base58' for legacy and p2sh-segwit witness type and 'bech32' for segwit

    :param witness_type: Witness type: legacy, p2sh-segwit or segwit
    :type witness_type: str

    :return str:
    """

    if witness_type == "segwit":
        return "bech32"
    elif witness_type in [None, "legacy", "p2sh-segwit"]:
        return "base58"
    else:
        raise ValueError("Unknown witness type %s" % witness_type)



# -*- coding: utf-8 -*-
#
#    fluxwallet - Python Cryptocurrency Library
#    BitGo Client
#    © 2017-2019 July - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import asyncio
import binascii
import logging
from collections.abc import Iterator
from contextlib import aclosing
from datetime import datetime
from decimal import Decimal

import httpx
from rich.pretty import pprint

from fluxwallet.config.config import FLUXWALLET_VERSION
from fluxwallet.main import MAX_TRANSACTIONS
from fluxwallet.services.baseclient import BaseClient, ClientError
from fluxwallet.transactions import BaseTransaction, FluxTransaction

_logger = logging.getLogger(__name__)

PROVIDERNAME = "flux"
LIMIT_TX = 49


class FluxClient(BaseClient):
    def __init__(self, network: str, base_url: str, denominator: int, *args):
        super().__init__(network, PROVIDERNAME, base_url, denominator, *args)

    def load_tx(self, tx: dict) -> FluxTransaction:
        """Load a transaction from the api into a python object

        Args:
            tx (dict): The result from API call

        Returns:
            Transaction: a parsed transaction
        """

        confirmations = tx.get("confirmations", 0)
        status = "unconfirmed"
        if confirmations:
            status = "confirmed"
        witness_type = "legacy"

        coinbase = bool(tx.get("isCoinBase", False))
        fee = tx.get("fees", 0)
        value_in = tx.get("valueIn", 0)

        t = FluxTransaction(
            locktime=tx["locktime"],
            version=tx["version"],
            network="flux",
            fee=fee,
            size=tx["size"],
            txid=tx["txid"],
            date=None if not confirmations else datetime.utcfromtimestamp(tx["time"]),
            confirmations=confirmations,
            block_height=tx["blockheight"],
            status=status,
            input_total=value_in,
            coinbase=coinbase,
            output_total=tx["valueOut"],
            witness_type=witness_type,
            expiry_height=tx["nExpiryHeight"],
        )

        if coinbase:
            t.add_input(prev_txid=b"\00" * 32, output_n=0, value=0)
        else:
            index_n = 0
            for ti in tx["vin"]:
                t.add_input(
                    prev_txid=binascii.unhexlify(ti["txid"]),
                    output_n=ti["vout"],
                    unlocking_script=ti["scriptSig"]["hex"],
                    index_n=index_n,
                    value=ti["valueSat"],
                    address=ti["addr"],
                    sequence=ti["sequence"],
                    strict=True,
                )
                index_n += 1

        for to in tx["vout"]:
            try:
                addr = to["scriptPubKey"]["addresses"][0]
            except KeyError:
                addr = ""

            t.add_output(
                value=int(Decimal(to["value"]) * 100000000),
                address=addr,
                lock_script=to["scriptPubKey"]["hex"],
                spent=bool(to["spentTxId"]),
                output_n=to["n"],
                spending_txid=to["spentTxId"],
                spending_index_n=to["spentIndex"],
                strict=True,
            )

        return t

    async def getutxos(
        self, address: str, after_txid: str = "", limit: int = MAX_TRANSACTIONS
    ):
        utxos = []

        query_params = {"address": address}
        res = await self.do_get("explorer/utxo", params=query_params)

        after_height = 0

        for utxo in res["data"]:
            if utxo["txid"] == after_txid:
                after_height = int(utxo["height"])

            # need to go look up the tx to get size / fee etc etc.
            utxos.append(
                {
                    "address": utxo["address"],
                    "txid": utxo["txid"],
                    "confirmations": utxo["confirmations"],
                    "output_n": utxo["vout"],
                    "input_n": 0,
                    "block_height": int(utxo["height"]),
                    # "fee": None,
                    # "size": 0,
                    "value": utxo["satoshis"],
                    "script": utxo["scriptPubKey"],
                }
            )
        after_tx_filter = (
            lambda x: x["block_height"] >= after_height and x["txid"] != after_txid
        )
        utxos = list(filter(after_tx_filter, utxos))

        return utxos[::-1][:limit]

    async def estimatefee(self, blocks):
        # Fix this
        return 3

    async def blockcount(self) -> int:
        # return self.compose_request("daemon/getblockcount")["data"]
        res: dict = await anext(self.do_get("daemon/getblockcount"))
        print(res)
        return res.get("data", 0)

    async def sendrawtransaction(self, rawtx: str) -> dict[str, dict]:
        # res = self.compose_request(
        #     "daemon/sendrawtransaction",
        #     post_data={"hexstring": rawtx},
        #     http_verb="post",
        # )
        res = await self.do_post(
            "daemon/sendrawtransaction", post_data={"hexstring": rawtx}
        )

        try:
            res.raise_for_status()
        except httpx.RequestError:
            return

        tx_res = res.json()

        print(tx_res)

        return {
            "txid": tx_res["data"],
        }

    # async def gettransaction(self, txid: str):
    #     # variables = {"txid": txid}
    #     # res = self.compose_request("daemon", "getrawtransaction", variables=variables)
    #     base = "https://explorer.runonflux.io/"
    #     res = self.compose_request(f"api/tx/{txid}", base=base)

    #     tx = self.load_tx(res)
    #     # pprint(tx.as_dict())
    #     return tx

    async def get_transactions(self, txids: list[str]):
        txs: list[dict] = []

        paths = [f"tx/{txid}" for txid in txids]

        tx_gen = self.do_get(paths, base_url="https://explorer.runonflux.io/api/")
        # {'status': 404, 'url': '/api/tx/', 'error': 'Not found'}
        async for tx_group in tx_gen:
            txs.extend(tx_group)

        txs = sorted(txs, key=lambda x: x["blockheight"])
        tx_objects = [self.load_tx(tx) for tx in txs]

        return tx_objects

    async def get_transactions_by_address(
        self, address: str, after_tx_index: int = 0, limit: int = 0
    ) -> Iterator[list[FluxTransaction]]:
        # get lastest txid and count
        params = {"from": 0, "to": 1}

        after_txid_index = None
        # https://explorer.runonflux.io/api/addrs/t1XvGeQCfYMhfagYb3GmbKRajNEjpPRZHkB/txs?from=0&to=1&noAsm=1&noScriptSig=1

        # don't use generator for single items, or just use for to get the single item so we don't have to use aclosing
        async with aclosing(
            self.do_get(
                f"/api/addrs/{address}/txs",
                base_url="https://explorer.runonflux.io",
                params=params,
            )
        ) as gen:
            res: dict = await anext(gen)

        txs: list[dict] = res.get("items", [])
        tx_count: int = res.get("totalItems", 0)

        if not txs:
            return

        if tx_count == 1:
            yield [self.load_tx(txs[0])]
            return

        # if after_txid:
        #     target_tx = next(filter(lambda x: x["txid"] == after_txid, txs), None)
        #     if target_tx:
        #         after_txid_index = txids.index(target_tx) + 1

        # from_tx = after_txid_index if after_txid_index else 0

        # this is always 0, newest
        from_tx = 0
        to_tx = (tx_count - after_tx_index) + 1

        tx_generator = self.do_get(
            paths=f"/api/addrs/{address}/txs",
            base_url="https://explorer.runonflux.io",
            pages=(from_tx, to_tx),
            chunksize=1,
        )

        # tx_generator = self.do_get(
        #     paths=["/api/txs"],
        #     base_url="https://explorer.runonflux.io",
        #     params=params,
        # )

        # total_count = 0
        async for batch in tx_generator:
            txs = []
            for result in batch:
                # total_count += len(result.get("txs", []))
                txs.extend(result.get("items"))
            yield [self.load_tx(tx) for tx in txs]

        # print("TOTAL COUNT", total_count)

        # target_blockheight = 0
        # if target_tx:
        #     target_blockheight = target_tx["blockheight"]

        # if target_blockheight:
        #     txs = list(
        #         filter(
        #             lambda x: x["blockheight"] >= target_blockheight
        #             and x["txid"] != after_txid,
        #             txs,
        #         )
        #     )
        #     txs = sorted(txs, key=lambda x: x["blockheight"])

    # async def get_pages_for_item(
    #     self,
    #     item: str,
    #     client: httpx.AsyncClient,
    #     endpoint: str,
    #     params: dict,
    #     pages: int,
    # ) -> list:
    #     results = []
    #     tasks = [
    #         client.send(
    #             client.build_request(
    #                 "GET",
    #                 endpoint,
    #                 params=params | {"pageNum": page},
    #             )
    #         )
    #         for page in range(2, pages + 1)
    #     ]

    #     to_retry = []
    #     for coro in asyncio.as_completed(tasks):
    #         try:
    #             result = await coro
    #         except httpx.RequestError as e:
    #             pprint(type(e))
    #             pprint(e.request)
    #             # retry
    #             # add to missing... then retry them all
    #             print(f"Failure... Adding {e.request.url} to be retried")
    #             to_retry.append(e.request.url)
    #         else:
    #             data: dict = result.json()

    #         results.extend(data.get(item))

    #     retries = [
    #         client.send(client.build_request("GET", page_url)) for page_url in to_retry
    #     ]
    #     # this is highly regarded
    #     for coro in asyncio.as_completed(retries):
    #         result = await coro
    #         data: dict = result.json()
    #         results.extend(data.get(item))

    #     return results

    async def do_post(
        self,
        endpoint: str,
        *,
        base_url: str | None = None,
        post_data: dict[str, str] | None = None,
    ) -> httpx.Response:
        if not base_url:
            base_url = self.base_url

        headers = {
            "User-Agent": f"fluxwallet/{FLUXWALLET_VERSION}",
            "Accept": "application/json",
        }
        transport = httpx.AsyncHTTPTransport(retries=5)

        async with httpx.AsyncClient(
            base_url=base_url, headers=headers, transport=transport
        ) as client:
            # need to retry if failure or do something
            return await client.post(endpoint, data=post_data)

    async def do_get(
        self,
        paths: list[str] | str,
        *,
        base_url: str | None = None,
        params: dict[str, str] | None = None,
        chunksize: int = 10,
        pages: int | tuple[int] | None = None,
    ) -> Iterator[list[dict]]:
        if not base_url:
            base_url = self.base_url

        headers = {
            "User-Agent": f"fluxwallet/{FLUXWALLET_VERSION}",
            "Accept": "application/json",
        }
        transport = httpx.AsyncHTTPTransport(retries=6)
        timeout = httpx.Timeout(10.0, connect=5.0)
        # pprint(paths)
        async with httpx.AsyncClient(
            base_url=base_url, headers=headers, transport=transport, timeout=timeout
        ) as client:
            # this doesn't mix in the client options (have to build request)
            # req = httpx.Request(http_verb, uri)
            if not params:
                params = {}

            single_result = False
            if isinstance(paths, str):
                paths = [paths]
                single_result = True

            if pages:
                if isinstance(pages, int):
                    tasks = [
                        client.get(
                            paths[0],
                            params=params | {"pageNum": page},
                        )
                        # we already have the first page
                        for page in range(2, pages + 1)
                    ]
                else:  # tuple (from, to)
                    # fix if response to_x < response size, also missing last group
                    single_result = False
                    from_tx, to_tx = pages
                    response_size = 20

                    if to_tx < response_size:
                        ranges = [(0, 20)]
                    else:
                        to_tx = response_size * round(to_tx / response_size)

                        ranges = list(range(from_tx, to_tx + 1, response_size))
                        ranges = [*zip(ranges, ranges[1::])]

                    tasks = [
                        client.get(
                            paths[0], params=params | {"from": range[0], "to": range[1]}
                        )
                        for range in ranges
                    ]

            else:
                tasks = [client.get(path, params=params) for path in paths]

            results = []
            to_retry = []
            count = 0

            for coro in asyncio.as_completed(tasks):
                try:
                    result = await coro
                except httpx.RequestError as e:
                    print("REQUEST ERROR")
                    to_retry.append(e.request.url)
                else:
                    result = result.json()
                    if single_result:
                        # print("HTTX RESULT")
                        # pprint(result)
                        yield result

                    results.append(result)

                    pages_total = result.get("pagesTotal", 0)

                    if not pages and pages_total:
                        async for result in self.do_get(
                            paths, base_url=base_url, params=params, pages=pages_total
                        ):
                            yield result

                    count += 1

                if count == chunksize:
                    yield results
                    count = 0
                    results = []

            if count > 0:
                yield results

            if to_retry:
                print(f"Warning: Retrying {len(to_retry)} endpoints")
                if len(to_retry) == 1:
                    print(f"Retrying: {to_retry[0]}")

                if single_result:
                    to_retry = str(to_retry[0])  # it's a URL object

                async for result in self.do_get(to_retry):
                    yield result



# -*- coding: utf-8 -*-
#
#    fluxwallet - Python Cryptocurrency Library
#
#    CLW - Command Line Wallet manager.
#    Create and manage fluxwallet legacy/segwit single and multisignatures wallets from the commandline
#
#    © 2019 November - 1200 Web Development <http://1200wd.com/>
#

import argparse
import ast
import asyncio
import sys
from pprint import pprint

from fluxwallet.keys import HDKey
from fluxwallet.main import FLUXWALLET_VERSION
from fluxwallet.mnemonic import Mnemonic
from fluxwallet.wallet import Wallet, WalletTransaction
from fluxwallet.wallet.errors import WalletError

# from fluxwallet.wallet import (
#     Wallet,
#     WalletError,
#     wallet_delete,
#     wallet_empty,
#     wallet_exists,
#     wallets_list,
# )
from fluxwallet.wallet.helpers import wallet_delete, wallet_empty, wallet_exists

# import logging

# from rich.logging import RichHandler

# FORMAT = "%(name)s: %(message)s"
# logging.basicConfig(
#     level="NOTSET",
#     format=FORMAT,
#     datefmt="[%X]",
#     handlers=[RichHandler(level=logging.INFO)],
# )
# logging.getLogger("aiosqlite").disabled = True
# logging.getLogger("httpx").disabled = True
# logging.getLogger("httpcore.http11").disabled = True

try:
    import pyqrcode

    QRCODES_AVAILABLE = True
except ImportError:
    QRCODES_AVAILABLE = False


DEFAULT_NETWORK = "bitcoin"


def parse_args():
    parser = argparse.ArgumentParser(description="fluxwallet command line wallet")
    parser.add_argument(
        "wallet_name",
        nargs="?",
        default="",
        help="Name of wallet to create or open. Used to store your all your wallet keys "
        "and will be printed on each paper wallet",
    )

    group_wallet = parser.add_argument_group("Wallet Actions")
    group_wallet.add_argument(
        "--wallet-remove",
        action="store_true",
        help="Name or ID of wallet to remove, all keys and transactions will be deleted",
    )
    group_wallet.add_argument(
        "--list-wallets",
        "-l",
        action="store_true",
        help="List all known wallets in fluxwallet database",
    )
    group_wallet.add_argument(
        "--wallet-info", "-w", action="store_true", help="Show wallet information"
    )
    group_wallet.add_argument(
        "--update-utxos",
        "-x",
        action="store_true",
        help="Update unspent transaction outputs (UTXO's) for this wallet",
    )
    group_wallet.add_argument(
        "--update-transactions",
        "-u",
        action="store_true",
        help="Update all transactions and UTXO's for this wallet",
    )
    group_wallet.add_argument(
        "--wallet-recreate",
        "-z",
        action="store_true",
        help="Delete all keys and transactions and recreate wallet, except for the masterkey(s)."
        " Use when updating fails or other errors occur. Please backup your database and "
        "masterkeys first.",
    )
    group_wallet.add_argument(
        "--receive",
        "-r",
        nargs="?",
        type=int,
        help="Show unused address to receive funds. Specify cosigner-id to generate address for "
        "specific cosigner. Default is -1 for own wallet",
        const=-1,
        metavar="COSIGNER_ID",
    )
    group_wallet.add_argument(
        "--generate-key",
        "-g",
        action="store_true",
        help="Generate a new masterkey, and show"
        " passphrase, WIF and public account key. Can be used to create a multisig wallet",
    )
    group_wallet.add_argument(
        "--export-private",
        "-e",
        action="store_true",
        help="Export private key for this wallet and exit",
    )
    group_wallet.add_argument(
        "--import-private", "-k", help="Import private key in this wallet"
    )

    group_wallet2 = parser.add_argument_group("Wallet Setup")
    group_wallet2.add_argument(
        "--passphrase",
        nargs="*",
        default=None,
        help="Passphrase to recover or create a wallet. Usually 12 or 24 words",
    )
    group_wallet2.add_argument(
        "--passphrase-strength",
        type=int,
        default=128,
        help="Number of bits for passphrase key. Default is 128, lower is not adviced but can "
        "be used for testing. Set to 256 bits for more future proof passphrases",
    )
    group_wallet2.add_argument(
        "--network",
        "-n",
        help="Specify 'bitcoin', 'litecoin', 'testnet' or other supported network",
    )
    group_wallet2.add_argument(
        "--database",
        "-d",
        help="URI of the database to use",
    )
    group_wallet2.add_argument(
        "--create-from-key",
        "-c",
        metavar="KEY",
        help="Create a new wallet from specified key",
    )
    group_wallet2.add_argument(
        "--create-multisig",
        "-m",
        nargs="*",
        metavar=".",
        help="[NUMBER_OF_SIGNATURES, NUMBER_OF_SIGNATURES_REQUIRED, [KEY1, KEY2, ... KEY3]]"
        "Specificy number of signatures followed by the number of signatures required and "
        "then a list of public or private keys for this wallet. Private keys will be "
        "created if not provided in key list."
        "\nExample, create a 2-of-2 multisig wallet and provide 1 key and create another "
        "key: -m 2 2 tprv8ZgxMBicQKsPd1Q44tfDiZC98iYouKRC2CzjT3HGt1yYw2zuX2awTotzGAZQ"
        "EAU9bi2M5MCj8iedP9MREPjUgpDEBwBgGi2C8eK5zNYeiX8 tprv8ZgxMBicQKsPeUbMS6kswJc11zgV"
        "EXUnUZuGo3bF6bBrAg1ieFfUdPc9UHqbD5HcXizThrcKike1c4z6xHrz6MWGwy8L6YKVbgJMeQHdWDp",
    )
    group_wallet2.add_argument(
        "--witness-type",
        "-y",
        metavar="WITNESS_TYPE",
        default=None,
        help="Witness type of wallet: lecacy (default), p2sh-segwit or segwit",
    )
    group_wallet2.add_argument(
        "--cosigner-id",
        "-s",
        type=int,
        default=0,
        help="Set this if wallet contains only public keys, more then one private key or if "
        "you would like to create keys for other cosigners.",
    )
    group_transaction = parser.add_argument_group("Transactions")
    group_transaction.add_argument(
        "--create-transaction",
        "-t",
        metavar=("ADDRESS_1", "AMOUNT_1"),
        help="Create transaction. Specify address followed by amount. Repeat for multiple "
        "outputs",
        nargs="*",
    )
    group_transaction.add_argument(
        "--sweep",
        metavar="ADDRESS",
        help="Sweep wallet, transfer all funds to specified address",
    )
    group_transaction.add_argument("--fee", "-f", type=int, help="Transaction fee")
    group_transaction.add_argument(
        "--message",
        metavar="MESSAGE",
        help="Adds a string to the transaction" "outputs",
    )
    group_transaction.add_argument(
        "--fee-per-kb",
        type=int,
        help="Transaction fee in sathosis (or smallest denominator) per kilobyte",
    )
    group_transaction.add_argument(
        "--push",
        "-p",
        action="store_true",
        help="Push created transaction to the network",
    )
    group_transaction.add_argument(
        "--import-tx",
        "-i",
        metavar="TRANSACTION",
        help="Import raw transaction hash or transaction dictionary in wallet and sign "
        "it with available key(s)",
    )
    group_transaction.add_argument(
        "--import-tx-file",
        "-a",
        metavar="FILENAME_TRANSACTION",
        help="Import transaction dictionary or raw transaction string from specified "
        "filename and sign it with available key(s)",
    )

    pa = parser.parse_args()
    if pa.receive and pa.create_transaction:
        parser.error("Please select receive or create transaction option not both")
    if pa.wallet_name:
        pa.wallet_info = True
    else:
        pa.list_wallets = True
    return pa


def get_passphrase(args):
    inp_passphrase = Mnemonic("english").generate(args.passphrase_strength)
    print("\nYour mnemonic private key sentence is: %s" % inp_passphrase)
    print(
        "\nPlease write down on paper and backup. With this key you can restore your wallet and all keys"
    )
    passphrase = inp_passphrase.split(" ")
    inp = input("\nType 'yes' if you understood and wrote down your key: ")
    if inp not in ["yes", "Yes", "YES"]:
        clw_exit("Exiting...")
    return passphrase


async def create_wallet(wallet_name, args, db_uri):
    if args.network is None:
        args.network = DEFAULT_NETWORK
    print("\nCREATE wallet '%s' (%s network)" % (wallet_name, args.network))
    if args.create_multisig:
        if not isinstance(args.create_multisig, list) or len(args.create_multisig) < 2:
            clw_exit(
                "Please enter multisig creation parameter in the following format: "
                "<number-of-signatures> <number-of-signatures-required> "
                "<key-0> <key-1> [<key-2> ... <key-n>]"
            )
        try:
            sigs_total = int(args.create_multisig[0])
        except ValueError:
            clw_exit(
                "Number of total signatures (first argument) must be a numeric value. %s"
                % args.create_multisig[0]
            )
        try:
            sigs_required = int(args.create_multisig[1])
        except ValueError:
            clw_exit(
                "Number of signatures required (second argument) must be a numeric value. %s"
                % args.create_multisig[1]
            )
        key_list = args.create_multisig[2:]
        keys_missing = sigs_total - len(key_list)
        assert keys_missing >= 0
        if keys_missing:
            print("Not all keys provided, creating %d additional key(s)" % keys_missing)
            for _ in range(keys_missing):
                passphrase = get_passphrase(args)
                passphrase = " ".join(passphrase)
                seed = Mnemonic().to_seed(passphrase).hex()
                key_list.append(HDKey.from_seed(seed, network=args.network))
        return await Wallet.create(
            wallet_name,
            key_list,
            sigs_required=sigs_required,
            network=args.network,
            cosigner_id=args.cosigner_id,
            db_uri=db_uri,
            witness_type=args.witness_type,
        )
    elif args.create_from_key:
        from fluxwallet.keys import get_key_format

        import_key = args.create_from_key
        kf = get_key_format(import_key)
        if kf["format"] == "wif_protected":
            password = input("Key password? ")
            import_key, _ = HDKey._bip38_decrypt(import_key, password)
        return Wallet.create(
            wallet_name,
            import_key,
            network=args.network,
            db_uri=db_uri,
            witness_type=args.witness_type,
        )
    else:
        passphrase = args.passphrase
        if passphrase is None:
            passphrase = get_passphrase(args)
        elif not passphrase:
            passphrase = input("Enter Passphrase: ")
        if not isinstance(passphrase, list):
            passphrase = passphrase.split(" ")
        elif len(passphrase) == 1:
            passphrase = passphrase[0].split(" ")
        if len(passphrase) < 12:
            clw_exit("Please specify passphrase with 12 words or more")
        passphrase = " ".join(passphrase)
        seed = Mnemonic().to_seed(passphrase).hex()
        hdkey = HDKey.from_seed(seed, network=args.network)
        return await Wallet.create(
            wallet_name,
            hdkey,
            network=args.network,
            witness_type=args.witness_type,
            db_uri=db_uri,
        )


async def create_transaction(wlt: Wallet, send_args, args):
    message = args.message if args.message else ""

    output_arr = []
    while send_args:
        if len(send_args) == 1:
            raise ValueError(
                "Invalid number of transaction input use <address1> <amount1> ... <address_n> <amount_n>"
            )

        try:
            amount = int(send_args[1])
        except ValueError:
            clw_exit("Amount must be a integer value: %s" % send_args[1])

        output_arr.append((send_args[0], amount))
        send_args = send_args[2:]

    return await wlt.transaction_create(
        output_arr=output_arr,
        network=args.network,
        fee=args.fee,
        min_confirms=0,
        message=message,
    )


def print_transaction(wt: WalletTransaction):
    tx_dict = {
        "network": wt.tx.network.name,
        "fee": wt.tx.fee,
        "raw": wt.tx.raw_hex(),
        "outputs": [{"address": o.address, "value": o.value} for o in wt.tx.outputs],
        "inputs": [
            {
                "prev_hash": i.prev_txid.hex(),
                "output_n": int.from_bytes(i.output_n, "big"),
                "address": i.address,
                "signatures": [
                    {
                        "signature": s.hex(),
                        "sig_der": s.as_der_encoded(as_hex=True),
                        "pub_key": s.public_key.public_hex,
                    }
                    for s in i.signatures
                ],
                "value": i.value,
            }
            for i in wt.tx.inputs
        ],
    }
    pprint(tx_dict)


def clw_exit(msg=None):
    if msg:
        print(msg)
    sys.exit()


async def main():
    loop = asyncio.get_event_loop()
    loop.slow_callback_duration = 40

    print("Command Line Wallet - fluxwallet %s\n" % FLUXWALLET_VERSION)
    # --- Parse commandline arguments ---
    args = parse_args()

    db_uri = args.database

    if args.generate_key:
        passphrase = get_passphrase(args)
        passphrase = " ".join(passphrase)
        seed = Mnemonic().to_seed(passphrase).hex()
        hdkey = HDKey.from_seed(seed, network=args.network)
        print(
            "Private Master key, to create multisig wallet on this machine:\n%s"
            % hdkey.wif_private()
        )
        print(
            "Public Master key, to share with other cosigner multisig wallets:\n%s"
            % hdkey.public_master(witness_type=args.witness_type, multisig=True).wif()
        )
        print("Network: %s" % hdkey.network.name)
        clw_exit()

    # List wallets, then exit
    if args.list_wallets:
        print("fluxwallet wallets:")
        for w in wallets_list(db_uri=db_uri):
            if "parent_id" in w and w["parent_id"]:
                continue
            print("[%d] %s (%s) %s" % (w["id"], w["name"], w["network"], w["owner"]))
        clw_exit()

    # Delete specified wallet, then exit
    if args.wallet_remove:
        if not await wallet_exists(args.wallet_name, db_uri=db_uri):
            clw_exit("Wallet '%s' not found" % args.wallet_name)
        inp = input(
            "\nWallet '%s' with all keys and will be removed, without private key it cannot be restored."
            "\nPlease retype exact name of wallet to proceed: " % args.wallet_name
        )
        if inp == args.wallet_name:
            if await wallet_delete(args.wallet_name, force=True, db_uri=db_uri):
                clw_exit("\nWallet %s has been removed" % args.wallet_name)
            else:
                clw_exit("\nError when deleting wallet")
        else:
            clw_exit("\nSpecified wallet name incorrect")

    wlt = None

    if (
        args.wallet_name
        and not args.wallet_name.isdigit()
        and not await wallet_exists(args.wallet_name, db_uri=db_uri)
    ):
        if (
            not args.create_from_key
            and input(
                "Wallet %s does not exist, create new wallet [yN]? " % args.wallet_name
            ).lower()
            != "y"
        ):
            clw_exit("Aborted")
        wlt = await create_wallet(args.wallet_name, args, db_uri)
        await wlt.new_key(network="bitcoin")

        args.wallet_info = True
    else:
        try:
            wlt = await Wallet.open(args.wallet_name, db_uri=db_uri)
            if args.passphrase is not None:
                print("WARNING: Using passphrase option for existing wallet ignored")
            if args.create_from_key is not None:
                print(
                    "WARNING: Using create_from_key option for existing wallet ignored"
                )
        except WalletError as e:
            clw_exit("Error: %s" % e.msg)

    if wlt is None:
        clw_exit("Could not open wallet %s" % args.wallet_name)

    if args.import_private:
        if wlt.import_key(args.import_private):
            clw_exit("Private key imported")
        else:
            clw_exit("Failed to import key")

    if args.wallet_recreate:
        await wallet_empty(args.wallet_name)
        print("Removed transactions and generated keys from this wallet")
    if args.update_utxos:
        wlt.utxos_update()
    if args.update_transactions:
        await wlt.scan(scan_gap_limit=5)

    if args.export_private:
        if wlt.scheme == "multisig":
            for w in wlt.cosigner:
                if w.main_key and w.main_key.is_private:
                    print(w.main_key.wif)
        elif not wlt.main_key or not wlt.main_key.is_private:
            print("No private key available for this wallet")
        else:
            print(wlt.main_key.wif)
        clw_exit()

    if args.network is None:
        args.network = wlt.network.name

    tx_import = None
    if args.import_tx_file:
        try:
            fn = args.import_tx_file
            f = open(fn, "r")
        except FileNotFoundError:
            clw_exit("File %s not found" % args.import_tx_file)
        try:
            tx_import = ast.literal_eval(f.read())
        except (ValueError, SyntaxError):
            tx_import = f.read()
    if args.import_tx:
        try:
            tx_import = ast.literal_eval(args.import_tx)
        except (ValueError, SyntaxError):
            tx_import = args.import_tx
    if tx_import:
        if isinstance(tx_import, dict):
            wt = wlt.transaction_import(tx_import)
        else:
            wt = wlt.transaction_import_raw(tx_import, network=args.network)
        wt.sign()
        if args.push:
            res = wt.send()
            if res:
                print("Transaction pushed to network. Transaction ID: %s" % wt.txid)
            else:
                print("Error creating transaction: %s" % wt.error)
        wt.info()
        print("Signed transaction:")
        print_transaction(wt)
        clw_exit()

    if args.receive:
        cosigner_id = args.receive
        if args.receive == -1:
            cosigner_id = None
        key = await wlt.get_key(network=args.network, cosigner_id=cosigner_id)
        print("Receive address: %s" % key.address)
        if QRCODES_AVAILABLE:
            qrcode = pyqrcode.create(key.address)
            print(qrcode.terminal())
        else:
            print("Install qr code module to show QR codes: pip install pyqrcode")
        clw_exit()
    if args.create_transaction == []:
        clw_exit("Missing arguments for --create-transaction/-t option")
    if args.create_transaction:
        if args.fee_per_kb:
            clw_exit("Fee-per-kb option not allowed with --create-transaction")
        try:
            wt = await create_transaction(wlt, args.create_transaction, args)
        except WalletError as e:
            clw_exit("Cannot create transaction: %s" % e.msg)
        wt.sign()
        print("Transaction created")
        wt.info()
        if args.push:
            await wt.send()
            if wt.pushed:
                print("Transaction pushed to network. Transaction ID: %s" % wt.txid)
            else:
                print("Error creating transaction: %s" % wt.error)
        else:
            print(
                "\nTransaction created but not sent yet. Transaction dictionary for export: "
            )
            print_transaction(wt)
        clw_exit()
    if args.sweep:
        if args.fee:
            clw_exit("Fee option not allowed with --sweep")
        offline = True
        print("Sweep wallet. Send all funds to %s" % args.sweep)
        if args.push:
            offline = False
        wt = wlt.sweep(
            args.sweep,
            offline=offline,
            network=args.network,
            fee_per_kb=args.fee_per_kb,
        )
        if not wt:
            clw_exit(
                "Error occurred when sweeping wallet: %s. Are UTXO's available and updated?"
                % wt
            )
        wt.info()
        if args.push:
            if wt.pushed:
                print("Transaction pushed to network. Transaction ID: %s" % wt.txid)
            elif not wt:
                print("Cannot sweep wallet, are UTXO's updated and available?")
            else:
                print("Error sweeping wallet: %s" % wt.error)
        else:
            print(
                "\nTransaction created but not sent yet. Transaction dictionary for export: "
            )
            print_transaction(wt)
        clw_exit()

    # print("Updating wallet")
    if args.network == "fluxwallet_test":
        wlt.utxos_update()
    print("Wallet info for %s" % wlt.name)
    await wlt.info()


def entrypoint():
    asyncio.run(main(), debug=True)


if __name__ == "__main__":
    asyncio.run(main(), debug=True)



# -*- coding: utf-8 -*-
#
#    fluxwallet - Python Cryptocurrency Library
#    CONFIG - Configuration settings
#    © 2022 October - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import configparser
import enum
import locale
import os
import platform
from datetime import datetime
from pathlib import Path

# General defaults
TYPE_TEXT = str
TYPE_INT = int
LOGLEVEL = "DEBUG"


# File locations
FW_CONFIG_FILE = ""
FW_INSTALL_DIR = Path(__file__).parents[1]
FW_DATA_DIR = ""
FW_DATABASE_DIR = ""
DEFAULT_DATABASE = None
DEFAULT_DATABASE_CACHE = None
FW_LOG_FILE = ""

# Main
ENABLE_fluxwallet_LOGGING = True
ALLOW_DATABASE_THREADS = None

# Services
TIMEOUT_REQUESTS = 5
# MAX_TRANSACTIONS = 20
MAX_TRANSACTIONS = 20000
BLOCK_COUNT_CACHE_TIME = 3
SERVICE_MAX_ERRORS = 3  # Fail service request when more then max errors occur for <SERVICE_MAX_ERRORS> providers

# Transactions
SCRIPT_TYPES_LOCKING = {
    # Locking scripts / scriptPubKey (Output)
    "p2pkh": ["OP_DUP", "OP_HASH160", "hash-20", "OP_EQUALVERIFY", "OP_CHECKSIG"],
    "p2sh": ["OP_HASH160", "hash-20", "OP_EQUAL"],
    "p2wpkh": ["OP_0", "hash-20"],
    "p2wsh": ["OP_0", "hash-32"],
    "p2tr": ["op_n", "hash-32"],
    "multisig": ["op_m", "multisig", "op_n", "OP_CHECKMULTISIG"],
    "p2pk": ["public_key", "OP_CHECKSIG"],
    "nulldata": ["OP_RETURN", "return_data"],
}

SCRIPT_TYPES_UNLOCKING = {
    # Unlocking scripts / scriptSig (Input)
    "sig_pubkey": ["signature", "SIGHASH_ALL", "public_key"],
    "p2sh_multisig": ["OP_0", "multisig", "redeemscript"],
    "p2sh_p2wpkh": ["OP_0", "OP_HASH160", "redeemscript", "OP_EQUAL"],
    "p2sh_p2wsh": ["OP_0", "push_size", "redeemscript"],
    "locktime_cltv": ["locktime_cltv", "OP_CHECKLOCKTIMEVERIFY", "OP_DROP"],
    "locktime_csv": ["locktime_csv", "OP_CHECKSEQUENCEVERIFY", "OP_DROP"],
    "signature": ["signature"],
}

SIGHASH_ALL = 1
SIGHASH_NONE = 2
SIGHASH_SINGLE = 3
SIGHASH_ANYONECANPAY = 80

SEQUENCE_LOCKTIME_DISABLE_FLAG = 1 << 31  # To enable sequence time locks
SEQUENCE_LOCKTIME_TYPE_FLAG = (
    1 << 22
)  # If set use timestamp based lock otherwise use block height
SEQUENCE_LOCKTIME_GRANULARITY = 9
SEQUENCE_LOCKTIME_MASK = 0x0000FFFF
SEQUENCE_ENABLE_LOCKTIME = 0xFFFFFFFE
SEQUENCE_REPLACE_BY_FEE = 0xFFFFFFFD

SIGNATURE_VERSION_STANDARD = 0
SIGNATURE_VERSION_SEGWIT = 1

# Mnemonics
DEFAULT_LANGUAGE = "english"

# Networks
DEFAULT_NETWORK = "bitcoin"
NETWORK_DENOMINATORS = {  # source: https://en.bitcoin.it/wiki/Units, https://en.wikipedia.org/wiki/Metric_prefix
    0.00000000000001: "µsat",
    0.00000000001: "msat",
    0.000000001: "n",
    0.00000001: "sat",
    0.0000001: "fin",
    0.000001: "µ",
    0.001: "m",
    0.01: "c",
    0.1: "d",
    1: "",
    10: "da",
    100: "h",
    1000: "k",
    1000000: "M",
    1000000000: "G",
    1000000000000: "T",
    1000000000000000: "P",
    1000000000000000000: "E",
    1000000000000000000000: "Z",
    1000000000000000000000000: "Y",
}

if os.name == "nt" and locale.getpreferredencoding() != "UTF-8":
    # TODO: Find a better windows hack
    import _locale

    _locale._getdefaultlocale = lambda *args: ["en_US", "utf8"]
elif locale.getpreferredencoding() != "UTF-8":
    raise EnvironmentError(
        "Locale is currently set to '%s'. "
        "This library needs the locale set to UTF-8 to function properly"
        % locale.getpreferredencoding()
    )

# Keys / Addresses
SUPPORTED_ADDRESS_ENCODINGS = ["base58", "bech32"]
ENCODING_BECH32_PREFIXES = ["bc", "tb", "ltc", "tltc", "tdash", "tdash", "blt"]
DEFAULT_WITNESS_TYPE = "legacy"
BECH32M_CONST = 0x2BC830A3

# Wallets
WALLET_KEY_STRUCTURES = [
    {
        "purpose": None,
        "script_type": "p2pkh",
        "witness_type": "legacy",
        "multisig": False,
        "encoding": "base58",
        "description": "Single key wallet with no hierarchical deterministic key structure",
        "key_path": ["m"],
    },
    {
        "purpose": 44,
        "script_type": "p2pkh",
        "witness_type": "legacy",
        "multisig": False,
        "encoding": "base58",
        "description": "Legacy wallet using pay-to-public-key-hash scripts",
        "key_path": [
            "m",
            "purpose'",
            "coin_type'",
            "account'",
            "change",
            "address_index",
        ],
    },
    {
        "purpose": 45,
        "script_type": "p2sh",
        "witness_type": "legacy",
        "multisig": True,
        "encoding": "base58",
        "description": "Legacy multisig wallet using pay-to-script-hash scripts",
        "key_path": ["m", "purpose'", "cosigner_index", "change", "address_index"],
    },
    {
        "purpose": 48,
        "script_type": "p2sh-p2wsh",
        "witness_type": "p2sh-segwit",
        "multisig": True,
        "encoding": "base58",
        "description": "Segwit multisig wallet using pay-to-wallet-script-hash scripts nested in p2sh scripts",
        "key_path": [
            "m",
            "purpose'",
            "coin_type'",
            "account'",
            "script_type'",
            "change",
            "address_index",
        ],
    },
    {
        "purpose": 48,
        "script_type": "p2wsh",
        "witness_type": "segwit",
        "multisig": True,
        "encoding": "bech32",
        "description": "Segwit multisig wallet using native segwit pay-to-wallet-script-hash scripts",
        "key_path": [
            "m",
            "purpose'",
            "coin_type'",
            "account'",
            "script_type'",
            "change",
            "address_index",
        ],
    },
    {
        "purpose": 49,
        "script_type": "p2sh-p2wpkh",
        "witness_type": "p2sh-segwit",
        "multisig": False,
        "encoding": "base58",
        "description": "Segwit wallet using pay-to-wallet-public-key-hash scripts nested in p2sh scripts",
        "key_path": [
            "m",
            "purpose'",
            "coin_type'",
            "account'",
            "change",
            "address_index",
        ],
    },
    {
        "purpose": 84,
        "script_type": "p2wpkh",
        "witness_type": "segwit",
        "multisig": False,
        "encoding": "bech32",
        "description": "Segwit multisig wallet using native segwit pay-to-wallet-public-key-hash scripts",
        "key_path": [
            "m",
            "purpose'",
            "coin_type'",
            "account'",
            "change",
            "address_index",
        ],
    },
    # {
    #     'purpose': 86,
    #     'script_type': 'p2tr',
    #     'witness_type': 'segwit',
    #     'multisig': False,
    #     'encoding': 'bech32',
    #     'description': 'Taproot single key wallet using P2TR transactions',
    #     'key_path': ["m", "purpose'", "coin_type'", "account'", "change", "address_index"]
    # },
]

# UNITTESTS
UNITTESTS_FULL_DATABASE_TEST = False

# CACHING
SERVICE_CACHING_ENABLED = True


def read_config():
    config = configparser.ConfigParser()

    def config_get(section, var, fallback, is_boolean=False):
        try:
            if is_boolean:
                val = config.getboolean(section, var, fallback=fallback)
            else:
                val = config.get(section, var, fallback=fallback)
            return val
        except Exception:
            return fallback

    global FW_INSTALL_DIR, FW_DATABASE_DIR, DEFAULT_DATABASE, FW_DATA_DIR, FW_CONFIG_FILE
    global ALLOW_DATABASE_THREADS, DEFAULT_DATABASE_CACHE
    global FW_LOG_FILE, LOGLEVEL, ENABLE_fluxwallet_LOGGING
    global TIMEOUT_REQUESTS, DEFAULT_LANGUAGE, DEFAULT_NETWORK, DEFAULT_WITNESS_TYPE
    global UNITTESTS_FULL_DATABASE_TEST, SERVICE_CACHING_ENABLED
    global SERVICE_MAX_ERRORS, BLOCK_COUNT_CACHE_TIME, MAX_TRANSACTIONS

    # Read settings from Configuration file provided in OS environment~/.fluxwallet/ directory
    config_file_name = os.environ.get("FW_CONFIG_FILE")
    if not config_file_name:
        FW_CONFIG_FILE = Path("~/.fluxwallet/config.ini").expanduser()
    else:
        FW_CONFIG_FILE = Path(config_file_name)
        if not FW_CONFIG_FILE.is_absolute():
            FW_CONFIG_FILE = Path(Path.home(), ".fluxwallet", FW_CONFIG_FILE)
        if not FW_CONFIG_FILE.exists():
            FW_CONFIG_FILE = Path(FW_INSTALL_DIR, "data", config_file_name)
        if not FW_CONFIG_FILE.exists():
            raise IOError(
                "fluxwallet configuration file not found: %s" % str(FW_CONFIG_FILE)
            )
    data = config.read(str(FW_CONFIG_FILE))
    FW_DATA_DIR = Path(
        config_get("locations", "data_dir", fallback="~/.fluxwallet")
    ).expanduser()
    # Database settings
    FW_DATABASE_DIR = Path(
        FW_DATA_DIR, config_get("locations", "database_dir", "database")
    )
    FW_DATABASE_DIR.mkdir(parents=True, exist_ok=True)
    default_databasefile = DEFAULT_DATABASE = config_get(
        "locations", "default_databasefile", fallback="fluxwallet.sqlite"
    )
    if not default_databasefile.startswith(
        "postgresql"
    ) or default_databasefile.startswith("mysql"):
        DEFAULT_DATABASE = str(Path(FW_DATABASE_DIR, default_databasefile))
    default_databasefile_cache = DEFAULT_DATABASE_CACHE = config_get(
        "locations", "default_databasefile_cache", fallback="fluxwallet_cache.sqlite"
    )
    if not default_databasefile_cache.startswith(
        "postgresql"
    ) or default_databasefile_cache.startswith("mysql"):
        DEFAULT_DATABASE_CACHE = str(Path(FW_DATABASE_DIR, default_databasefile_cache))
    ALLOW_DATABASE_THREADS = config_get(
        "common", "allow_database_threads", fallback=True, is_boolean=True
    )
    SERVICE_CACHING_ENABLED = config_get(
        "common", "service_caching_enabled", fallback=True, is_boolean=True
    )

    # Log settings
    ENABLE_fluxwallet_LOGGING = config_get(
        "logs", "enable_fluxwallet_logging", fallback=True, is_boolean=True
    )
    FW_LOG_FILE = Path(
        FW_DATA_DIR, config_get("logs", "log_file", fallback="fluxwallet.log")
    )
    FW_LOG_FILE.parent.mkdir(parents=True, exist_ok=True)
    LOGLEVEL = config_get("logs", "loglevel", fallback=LOGLEVEL)

    # Service settings
    TIMEOUT_REQUESTS = int(
        config_get("common", "timeout_requests", fallback=TIMEOUT_REQUESTS)
    )
    SERVICE_MAX_ERRORS = int(
        config_get("common", "service_max_errors", fallback=SERVICE_MAX_ERRORS)
    )
    MAX_TRANSACTIONS = int(
        config_get("common", "max_transactions", fallback=MAX_TRANSACTIONS)
    )
    BLOCK_COUNT_CACHE_TIME = int(
        config_get("common", "block_count_cache_time", fallback=BLOCK_COUNT_CACHE_TIME)
    )

    # Other settings
    DEFAULT_LANGUAGE = config_get(
        "common", "default_language", fallback=DEFAULT_LANGUAGE
    )
    DEFAULT_NETWORK = config_get("common", "default_network", fallback=DEFAULT_NETWORK)
    DEFAULT_WITNESS_TYPE = config_get(
        "common", "default_witness_type", fallback=DEFAULT_WITNESS_TYPE
    )

    full_db_test = os.environ.get("UNITTESTS_FULL_DATABASE_TEST")
    if full_db_test:
        if full_db_test in [1, True, "True", "true", "TRUE"]:
            UNITTESTS_FULL_DATABASE_TEST = True

    if not data:
        return False
    return True


# Copy data and settings to default settings directory if install.log is not found
def initialize_lib():
    global FW_INSTALL_DIR, FW_DATA_DIR, FLUXWALLET_VERSION
    instlogfile = Path(FW_DATA_DIR, "install.log")
    if instlogfile.exists():
        return

    with instlogfile.open("w") as f:
        install_message = (
            "fluxwallet installed, check further logs in fluxwallet.log\n\n"
            "If you remove this file all settings will be reset again to the default settings. "
            "This might be usefull after an update or when problems occur.\n\n"
            "Installation parameters. Include this parameters when reporting bugs and issues:\n"
            "fluxwallet version: %s\n"
            "Installation date : %s\n"
            "Python            : %s\n"
            "Compiler          : %s\n"
            "Build             : %s\n"
            "OS Version        : %s\n"
            "Platform          : %s\n"
            % (
                FLUXWALLET_VERSION,
                datetime.now().isoformat(),
                platform.python_version(),
                platform.python_compiler(),
                platform.python_build(),
                platform.version(),
                platform.platform(),
            )
        )
        f.write(install_message)

    # Copy data and settings file
    from shutil import copyfile

    for file in Path(FW_INSTALL_DIR, "data").iterdir():
        if file.suffix not in [".ini", ".json"]:
            continue
        copyfile(str(file), Path(FW_DATA_DIR, file.name))


# Initialize library
read_config()

with open(Path(FW_INSTALL_DIR, "config/VERSION"), "r") as f:
    FLUXWALLET_VERSION = f.read().strip()

initialize_lib()



import struct
from dataclasses import dataclass
from hashlib import blake2b
from io import BytesIO
from typing import BinaryIO

from fluxwallet.encoding import read_varbyteint

MAX_MONEY = 21000000 * 100000000
TX_EXPIRY_HEIGHT_THRESHOLD = 500000000

OVERWINTER_VERSION_GROUP_ID = 0x03C4827
OVERWINTER_TX_VERSION = 3

SAPLING_VERSION_GROUP_ID = 0x892F2085
SAPLING_TX_VERSION = 4

OP_DUP = 0x76
OP_HASH160 = 0xA9
OP_EQUAL = 0x87
OP_EQUALVERIFY = 0x88
OP_CHECKSIG = 0xAC
OP_RETURN = 0x6A

MAX_COMPACT_SIZE = 0x2000000

SIGHASH_ALL = 1
SIGHASH_NONE = 2
SIGHASH_SINGLE = 3
SIGHASH_ANYONECANPAY = 0x80


def write_compact_size(n, allow_u64=False):
    assert allow_u64 or n <= MAX_COMPACT_SIZE
    if n < 253:
        return struct.pack("B", n)
    elif n <= 0xFFFF:
        return struct.pack("B", 253) + struct.pack("<H", n)
    elif n <= 0xFFFFFFFF:
        return struct.pack("B", 254) + struct.pack("<I", n)
    else:
        return struct.pack("B", 255) + struct.pack("<Q", n)


@dataclass
class Script:
    _script: bytes = b""

    def __bool__(self):
        return bool(self._script)

    def __bytes__(self):
        return write_compact_size(len(self._script)) + self._script

    @staticmethod
    def encode(operations: list) -> bytes:
        encoded = b""
        for op in operations:
            if isinstance(op, int):
                encoded += op.to_bytes(1, "big")
            else:
                encoded += write_compact_size(len(op)) + op
        return encoded

    @classmethod
    def from_bytes(cls, b: bytes):
        try:
            b: BinaryIO = BytesIO(b)
        except TypeError:
            script = cls()
            script._script = b
            return script

        op = b.read(1)
        op = int.from_bytes(op, "little")
        match op:
            case x if x == OP_RETURN:
                size = read_varbyteint(b)
                data = b.read(size)
                msg = MessageScript(data.decode())
            case x if x == OP_HASH160:
                # p2sh
                size = read_varbyteint(b)
                script_hash = b.read(size)
                op_equal = int.from_bytes(b.read(1), "little")
                assert op_equal == OP_EQUAL
                msg = P2SHScript(script_hash)
            case x if x == OP_DUP:
                # p2pkh
                hash160 = int.from_bytes(b.read(1), "little")
                assert hash160 == OP_HASH160
                size = read_varbyteint(b)
                pubkey_hash = b.read(size)
                op_equalverify = int.from_bytes(b.read(1), "little")
                assert op_equalverify == OP_EQUALVERIFY
                op_checksig = int.from_bytes(b.read(1), "little")
                assert op_checksig == OP_CHECKSIG
                msg = P2PKHScript(pubkey_hash)
            case _:
                # raise ValueError("Unknown script")
                script = cls()
                script._script = b.getvalue()
                msg = script

        return msg

    def raw(self):
        return self._script


class P2PKHScript(Script):
    def __init__(self, pubkey_hash: bytes):
        self._script = self.encode(
            [OP_DUP, OP_HASH160, pubkey_hash, OP_EQUALVERIFY, OP_CHECKSIG]
        )


class P2SHScript(Script):
    def __init__(self, script_hash: bytes):
        self._script = self.encode([OP_HASH160, script_hash, OP_EQUAL])


class MessageScript(Script):
    def __init__(self, message: str):
        self._script = self.encode([OP_RETURN, message.encode("utf-8")])


@dataclass
class OutPoint:
    txid: bytes
    n: int

    def __bytes__(self):
        return self.txid + struct.pack("<I", self.n)


@dataclass
class TxIn:
    prevout: OutPoint
    scriptSig: Script
    nSequence: int

    def __bytes__(self):
        return (
            bytes(self.prevout)
            + bytes(self.scriptSig)
            + struct.pack("<I", self.nSequence)
        )


@dataclass
class TxOut:
    nValue: int
    script: Script

    def __bytes__(self):
        return struct.pack("<Q", self.nValue) + bytes(self.script)


class SaplingTx:
    def __init__(
        self,
        version,
        vin: list[TxIn] = [],
        vout: list[TxOut] = [],
        nLockTime: int = 0,
        nExpiryHeight: int = 0,
        valuebalance: int = 0,
        shielded_spends: list = [],
        shielded_outputs: list = [],
    ):
        if version == OVERWINTER_TX_VERSION:
            self.fOverwintered = True
            self.nVersionGroupId = OVERWINTER_VERSION_GROUP_ID
            self.nVersion = OVERWINTER_TX_VERSION
        elif version == SAPLING_TX_VERSION:
            self.fOverwintered = True
            self.nVersionGroupId = SAPLING_VERSION_GROUP_ID
            self.nVersion = SAPLING_TX_VERSION
        else:
            raise Exception(f"Version: {version} not supported for SaplingTx")

        self.vin = []
        for tx in vin:
            self.vin.append(tx)

        self.vout = []
        for tx in vout:
            self.vout.append(tx)

        self.nLockTime = nLockTime
        self.nExpiryHeight = nExpiryHeight % TX_EXPIRY_HEIGHT_THRESHOLD
        if self.nVersion >= SAPLING_TX_VERSION:
            self.valueBalance = valuebalance % (MAX_MONEY + 1)

        self.vShieldedSpends = []
        self.vShieldedOutputs = []
        if self.nVersion >= SAPLING_TX_VERSION:
            for spend in shielded_spends:
                self.vShieldedSpends.append(spend)
            for output in shielded_outputs:
                self.vShieldedOutputs.append(output)

        self.vJoinSplit = []

    def __bytes__(self):
        ret = b""
        ret += struct.pack("<I", self.version_bytes())
        if self.fOverwintered:
            ret += struct.pack("<I", self.nVersionGroupId)

        isOverwinterV3 = (
            self.fOverwintered
            and self.nVersionGroupId == OVERWINTER_VERSION_GROUP_ID
            and self.nVersion == OVERWINTER_TX_VERSION
        )

        isSaplingV4 = (
            self.fOverwintered
            and self.nVersionGroupId == SAPLING_VERSION_GROUP_ID
            and self.nVersion == SAPLING_TX_VERSION
        )

        ret += write_compact_size(len(self.vin))
        for x in self.vin:
            ret += bytes(x)

        ret += write_compact_size(len(self.vout))
        for x in self.vout:
            ret += bytes(x)

        ret += struct.pack("<I", self.nLockTime)
        if isOverwinterV3 or isSaplingV4:
            ret += struct.pack("<I", self.nExpiryHeight)

        if isSaplingV4:
            ret += struct.pack("<Q", self.valueBalance)
            ret += write_compact_size(len(self.vShieldedSpends))
            for desc in self.vShieldedSpends:
                ret += bytes(desc)
            ret += write_compact_size(len(self.vShieldedOutputs))
            for desc in self.vShieldedOutputs:
                ret += bytes(desc)

        if self.nVersion >= 2:
            ret += write_compact_size(len(self.vJoinSplit))
            for jsdesc in self.vJoinSplit:
                ret += bytes(jsdesc)
            if len(self.vJoinSplit) > 0:
                ret += self.joinSplitPubKey
                ret += self.joinSplitSig

        if isSaplingV4 and not (
            len(self.vShieldedSpends) == 0 and len(self.vShieldedOutputs) == 0
        ):
            ret += self.bindingSig

        return ret

    def __repr__(self):
        resp = ""
        for k, v in self.__dict__.items():
            resp += f"{k}: {v}\n"
        return resp

    def version_bytes(self):
        return self.nVersion | (1 << 31 if self.fOverwintered else 0)

    def getHashPrevouts(self, person=b"ZcashPrevoutHash"):
        digest = blake2b(digest_size=32, person=person)

        for x in self.vin:
            digest.update(bytes(x.prevout))
        return digest.digest()

    def getHashSequence(self, person=b"ZcashSequencHash"):
        digest = blake2b(digest_size=32, person=person)

        for x in self.vin:
            digest.update(struct.pack("<I", x.nSequence))
        return digest.digest()

    def getHashOutputs(self, person=b"ZcashOutputsHash"):
        digest = blake2b(digest_size=32, person=person)
        for x in self.vout:
            digest.update(bytes(x))
        return digest.digest()

    def getHashJoinSplits(self):
        digest = blake2b(digest_size=32, person=b"ZcashJSplitsHash")
        for jsdesc in self.vJoinSplit:
            digest.update(bytes(jsdesc))
        digest.update(self.joinSplitPubKey)
        return digest.digest()

    def getHashShieldedSpends(self):
        digest = blake2b(digest_size=32, person=b"ZcashSSpendsHash")
        for desc in self.vShieldedSpends:
            # We don't pass in serialized form of desc as spendAuthSig is not part of the hash
            digest.update(bytes(desc.cv))
            digest.update(bytes(desc.anchor))
            digest.update(desc.nullifier)
            digest.update(bytes(desc.rk))
            digest.update(bytes(desc.proof))
        return digest.digest()

    def getHashShieldedOutputs(self):
        digest = blake2b(digest_size=32, person=b"ZcashSOutputHash")
        for desc in self.vShieldedOutputs:
            digest.update(bytes(desc))
        return digest.digest()

    def signature_hash(
        self,
        script_code: bytes,
        value: int,
        input_n: int,
        nIn: int,
        hash_type: str = SIGHASH_ALL,
    ) -> bytes:
        consensusBranchId = 0x76B809BB  # Sapling
        if input_n == None:
            return b""

        # this is zip243
        hashPrevouts = b"\x00" * 32
        hashSequence = b"\x00" * 32
        hashOutputs = b"\x00" * 32
        hashJoinSplits = b"\x00" * 32
        hashShieldedSpends = b"\x00" * 32
        hashShieldedOutputs = b"\x00" * 32

        if not (hash_type & SIGHASH_ANYONECANPAY):
            hashPrevouts = self.getHashPrevouts()

        if (
            (not (hash_type & SIGHASH_ANYONECANPAY))
            and (hash_type & 0x1F) != SIGHASH_SINGLE
            and (hash_type & 0x1F) != SIGHASH_NONE
        ):
            hashSequence = self.getHashSequence()

        if (hash_type & 0x1F) != SIGHASH_SINGLE and (hash_type & 0x1F) != SIGHASH_NONE:
            hashOutputs = self.getHashOutputs()
        elif (hash_type & 0x1F) == SIGHASH_SINGLE and 0 <= nIn and nIn < len(self.vout):
            digest = blake2b(digest_size=32, person=b"ZcashOutputsHash")
            digest.update(bytes(self.vout[nIn]))
            hashOutputs = digest.digest()

        if len(self.vJoinSplit) > 0:
            hashJoinSplits = self.getHashJoinSplits()

        if len(self.vShieldedSpends) > 0:
            hashShieldedSpends = self.getHashShieldedSpends()

        if len(self.vShieldedOutputs) > 0:
            hashShieldedOutputs = self.getHashShieldedOutputs()

        # print("hashPrevouts", binascii.hexlify(hashPrevouts))
        # print("hashSequence", binascii.hexlify(hashSequence))
        # print("hashOutputs", binascii.hexlify(hashOutputs))
        # print("hashJoinSplits", binascii.hexlify(hashJoinSplits))
        # print("hashShieldedSpends", binascii.hexlify(hashShieldedSpends))
        # print("hashShieldedOutputs", binascii.hexlify(hashShieldedOutputs))
        # print("concensusbranch", consensusBranchId)

        digest = blake2b(
            digest_size=32,
            person=b"ZcashSigHash" + struct.pack("<I", consensusBranchId),
        )

        digest.update(struct.pack("<I", self.version_bytes()))
        digest.update(struct.pack("<I", self.nVersionGroupId))
        digest.update(hashPrevouts)
        digest.update(hashSequence)
        digest.update(hashOutputs)
        digest.update(hashJoinSplits)
        digest.update(hashShieldedSpends)
        digest.update(hashShieldedOutputs)
        digest.update(struct.pack("<I", self.nLockTime))
        digest.update(struct.pack("<I", self.nExpiryHeight))
        digest.update(struct.pack("<Q", self.valueBalance))
        digest.update(struct.pack("<I", hash_type))

        digest.update(bytes(self.vin[input_n].prevout))
        digest.update(write_compact_size(len(script_code)) + script_code)
        digest.update(struct.pack("<Q", value))

        digest.update(struct.pack("<I", self.vin[input_n].nSequence))

        return digest.digest()



# -*- coding: utf-8 -*-
#
#    fluxwallet - Python Cryptocurrency Library
#    Blockchair client
#    © 2018-2019 July - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
from datetime import datetime

try:
    from datetime import timezone
except Exception:
    pass
from fluxwallet.encoding import EncodingError, varstr
from fluxwallet.keys import Address, deserialize_address
from fluxwallet.main import MAX_TRANSACTIONS
from fluxwallet.services.baseclient import BaseClient, ClientError
from fluxwallet.transactions import BitcoinTransaction

_logger = logging.getLogger(__name__)

PROVIDERNAME = "blockchair"
REQUEST_LIMIT = 100


class BlockChairClient(BaseClient):
    def __init__(self, network, base_url, denominator, *args):
        super(self.__class__, self).__init__(
            network, PROVIDERNAME, base_url, denominator, *args
        )

    def compose_request(
        self,
        command,
        query_vars=None,
        variables=None,
        data=None,
        offset=0,
        limit=REQUEST_LIMIT,
        method="get",
    ):
        url_path = ""
        if not variables:
            variables = {}
        if command not in ["stats", "mempool"]:
            variables.update({"limit": limit})
        if offset:
            variables.update({"offset": offset})
        if command:
            url_path += command
        if data:
            if url_path[-1:] != "/":
                url_path += "/"
            url_path += data
        if query_vars:
            varstr = ",".join(["%s(%s)" % (qv, query_vars[qv]) for qv in query_vars])
            variables.update({"q": varstr})
        return self.request(url_path, variables, method=method)

    def getbalance(self, addresslist):
        balance = 0
        for address in addresslist:
            res = self.compose_request("dashboards/address/", data=address)
            balance += int(res["data"][address]["address"]["balance"])
        return balance

    def getutxos(self, address, after_txid="", limit=MAX_TRANSACTIONS):
        utxos = []
        offset = 0
        while True:
            res = self.compose_request(
                "outputs", {"recipient": address, "is_spent": "false"}, offset=offset
            )
            if len(res["data"]) == REQUEST_LIMIT:
                raise ClientError(
                    "Blockchair returned more then maximum of %d data rows"
                    % REQUEST_LIMIT
                )
            current_block = res["context"]["state"]
            for utxo in res["data"][::-1]:
                if utxo["is_spent"]:
                    continue
                if utxo["transaction_hash"] == after_txid:
                    utxos = []
                    continue
                utxos.append(
                    {
                        "address": address,
                        "txid": utxo["transaction_hash"],
                        "confirmations": current_block - utxo["block_id"],
                        "output_n": utxo["index"],
                        "input_n": 0,
                        "block_height": utxo["block_id"],
                        "fee": None,
                        "size": 0,
                        "value": utxo["value"],
                        "script": utxo["script_hex"],
                        "date": datetime.strptime(utxo["time"], "%Y-%m-%d %H:%M:%S"),
                    }
                )
            if not len(res["data"]) or len(res["data"]) < REQUEST_LIMIT:
                break
            offset += REQUEST_LIMIT
        return utxos[:limit]

    def gettransaction(self, tx_id):
        res = self.compose_request("dashboards/transaction/", data=tx_id)

        tx = res["data"][tx_id]["transaction"]
        confirmations = (
            0 if tx["block_id"] <= 0 else res["context"]["state"] - tx["block_id"]
        )
        status = "unconfirmed"
        if confirmations:
            status = "confirmed"
        witness_type = "legacy"
        if tx["has_witness"]:
            witness_type = "segwit"
        input_total = tx["input_total"]
        t = BitcoinTransaction(
            locktime=tx["lock_time"],
            version=tx["version"],
            network=self.network,
            fee=tx["fee"],
            size=tx["size"],
            txid=tx["hash"],
            date=None
            if not confirmations
            else datetime.strptime(tx["time"], "%Y-%m-%d %H:%M:%S"),
            confirmations=confirmations,
            block_height=tx["block_id"] if tx["block_id"] > 0 else None,
            status=status,
            input_total=input_total,
            coinbase=tx["is_coinbase"],
            output_total=tx["output_total"],
            witness_type=witness_type,
        )
        index_n = 0
        if not res["data"][tx_id]["inputs"]:
            # This is a coinbase transaction, add input
            t.add_input(prev_txid=b"\00" * 32, output_n=0, value=0)

        for ti in res["data"][tx_id]["inputs"]:
            if ti["spending_witness"]:
                # witnesses = b"".join([varstr(bytes.fromhex(x)) for x in ti['spending_witness'].split(",")])
                witnesses = ti["spending_witness"].split(",")
                address = Address.parse(ti["recipient"])
                if address.script_type == "p2sh":
                    witness_type = "p2sh-segwit"
                else:
                    witness_type = "segwit"
                t.add_input(
                    prev_txid=ti["transaction_hash"],
                    output_n=ti["index"],
                    unlocking_script=ti["spending_signature_hex"],
                    witnesses=witnesses,
                    index_n=index_n,
                    value=ti["value"],
                    address=address,
                    witness_type=witness_type,
                    sequence=ti["spending_sequence"],
                    strict=self.strict,
                )
            else:
                t.add_input(
                    prev_txid=ti["transaction_hash"],
                    output_n=ti["index"],
                    unlocking_script=ti["spending_signature_hex"],
                    index_n=index_n,
                    value=ti["value"],
                    address=ti["recipient"],
                    unlocking_script_unsigned=ti["script_hex"],
                    sequence=ti["spending_sequence"],
                    strict=self.strict,
                )
            index_n += 1
        for to in res["data"][tx_id]["outputs"]:
            try:
                deserialize_address(to["recipient"], network=self.network.name)
                addr = to["recipient"]
            except EncodingError:
                addr = ""
            t.add_output(
                value=to["value"],
                address=addr,
                lock_script=to["script_hex"],
                spent=to["is_spent"],
                output_n=to["index"],
                spending_txid=to["spending_transaction_hash"],
                spending_index_n=to["spending_index"],
                strict=self.strict,
            )
        return t

    def gettransactions(self, address, after_txid="", limit=MAX_TRANSACTIONS):
        txids = []
        offset = 0
        while True:
            res = self.compose_request(
                "dashboards/address/", data=address, offset=offset
            )
            addr = res["data"][address]
            if not addr["transactions"]:
                break
            txids = addr["transactions"][::-1] + txids
            offset += 50
            if len(txids) > limit:
                break
        if after_txid:
            txids = txids[txids.index(after_txid) + 1 :]
        txs = []
        for txid in txids[:limit]:
            txs.append(self.gettransaction(txid))
        return txs

    def getrawtransaction(self, txid):
        res = self.compose_request("raw/transaction", data=txid)
        return res["data"][txid]["raw_transaction"]

    def sendrawtransaction(self, rawtx):
        res = self.compose_request(
            "push/transaction", variables={"data": rawtx}, method="post"
        )
        return {"txid": res["data"]["transaction_hash"], "response_dict": res}

    def estimatefee(self, blocks):
        return (
            self.compose_request("stats")["data"][
                "suggested_transaction_fee_per_byte_sat"
            ]
            * 1000
        )

    def blockcount(self):
        """
        Get latest block number: The block number of last block in longest chain on the blockchain

        :return int:
        """
        res = self.compose_request("stats")
        return res["context"]["state"]

    def mempool(self, txid=""):
        if txid:
            res = self.compose_request("mempool", {"hash": txid}, data="transactions")
        else:
            res = self.compose_request("mempool", data="transactions")
        return [tx["hash"] for tx in res["data"] if "hash" in tx]

    def getblock(self, blockid, parse_transactions, page, limit):
        if limit > 100:
            limit = 100
        res = self.compose_request(
            "dashboards/block/",
            data=str(blockid),
            offset=(page - 1) * limit,
            limit=limit,
        )
        bd = res["data"][str(blockid)]["block"]
        txids = res["data"][str(blockid)]["transactions"]
        if parse_transactions:
            txs = []
            for txid in txids:
                txs.append(self.gettransaction(txid))
        else:
            txs = txids

        block = {
            "bits": bd["bits"],
            "depth": None,
            "block_hash": bd["hash"],
            "height": bd["id"],
            "merkle_root": bd["merkle_root"],
            "nonce": bd["nonce"],
            "prev_block": b"",
            "time": int(
                datetime.strptime(bd["time"], "%Y-%m-%d %H:%M:%S")
                .replace(tzinfo=timezone.utc)
                .timestamp()
            ),
            "tx_count": bd["transaction_count"],
            "txs": txs,
            "version": bd["version"],
            "page": page,
            "pages": None
            if not limit
            else int(bd["transaction_count"] // limit)
            + (bd["transaction_count"] % limit > 0),
            "limit": limit,
        }
        return block

    def getrawblock(self, blockid):
        res = self.compose_request("raw/block/", data=str(blockid))
        rb = res["data"][str(blockid)]["raw_block"]
        return rb

    def isspent(self, txid, output_n):
        t = self.gettransaction(txid)
        return 1 if t.outputs[output_n].spent else 0

    def getinfo(self):
        info = self.compose_request("stats")["data"]
        return {
            "blockcount": info["best_block_height"],
            "chain": "",
            "difficulty": int(float(info["difficulty"])),
            "hashrate": int(info["hashrate_24h"]),
            "mempool_size": int(info["mempool_transactions"]),
        }



from __future__ import annotations

import pickle
import time
from itertools import groupby
from operator import itemgetter
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from fluxwallet.wallet import Wallet

from rich.pretty import pprint
from sqlalchemy import select
from sqlalchemy.exc import IntegrityError

from fluxwallet.db_new import (
    DbKey,
    DbTransaction,
    DbTransactionInput,
    DbTransactionOutput,
)

# from fluxwallet.db_new import
from fluxwallet.encoding import *
from fluxwallet.keys import Address, HDKey, check_network_and_key
from fluxwallet.mnemonic import Mnemonic
from fluxwallet.networks import Network
from fluxwallet.scripts import Script
from fluxwallet.services.services import Service
from fluxwallet.transactions import (
    BaseTransaction,
    BitcoinTransaction,
    FluxTransaction,
    Input,
    Output,
)
from fluxwallet.values import Value, value_to_satoshi

_logger = logging.getLogger(__name__)


class WalletTransaction:
    """
    Used as attribute of Wallet class. Child of Transaction object with extra reference to
    wallet and database object.

    All WalletTransaction items are stored in a database
    """

    def __init__(
        self, hdwallet: Wallet, account_id: int | None = None, *args, **kwargs
    ):
        """
        Initialize WalletTransaction object with reference to a Wallet object

        :param hdwallet: Wallet object, wallet name or ID
        :type hdWallet: HDwallet, str, int
        :param account_id: Account ID
        :type account_id: int
        :param args: Arguments for HDWallet parent class
        :type args: args
        :param kwargs: Keyword arguments for Wallet parent class
        :type kwargs: kwargs
        """
        # assert isinstance(hdwallet, Wallet)

        self.hdwallet = hdwallet
        self.account_id = account_id

        self.pushed = False
        self.error = None
        self.response_dict = None

        self.account_id = (
            self.hdwallet.default_account_id if account_id is None else account_id
        )

        witness_type = "legacy"
        if hdwallet.witness_type in ["segwit", "p2sh-segwit"]:
            witness_type = "segwit"

        # should set this in networks or something
        if self.hdwallet.network == "flux":
            self.__class__ = type(
                "WalletTransaction", (WalletTransaction, FluxTransaction), {}
            )
        else:
            self.__class__ = type(
                "WalletTransaction", (WalletTransaction, Transaction), {}
            )

        super().__init__(witness_type=witness_type, *args, **kwargs)

    @classmethod
    async def create(cls, hdwallet: Wallet, *args, **kwargs) -> WalletTransaction:
        self = WalletTransaction(hdwallet, *args, **kwargs)

        addresslist = await hdwallet.addresslist()

        self.outgoing_tx = bool(
            [i.address for i in self.inputs if i.address in addresslist]
        )
        self.incoming_tx = bool(
            [o.address for o in self.outputs if o.address in addresslist]
        )
        return self

    def __repr__(self):
        return (
            "<WalletTransaction(input_count=%d, output_count=%d, status=%s, network=%s)>"
            % (len(self.inputs), len(self.outputs), self.status, self.network.name)
        )

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        self_dict = self.__dict__
        for k, v in self_dict.items():
            if k != "hdwallet":
                setattr(result, k, deepcopy(v, memo))
        result.hdwallet = self.hdwallet
        return result

    @classmethod
    async def from_transaction(cls, hdwallet, t: BaseTransaction):
        """
        Create WalletTransaction object from Transaction object

        :param hdwallet: Wallet object, wallet name or ID
        :type hdwallet: HDwallet, str, int
        :param t: Specify Transaction object
        :type t: Transaction

        :return WalletClass:
        """

        return await cls.create(
            hdwallet=hdwallet,
            inputs=t.inputs,
            outputs=t.outputs,
            locktime=t.locktime,
            version=t.version,
            network=t.network.name,
            fee=t.fee,
            fee_per_kb=t.fee_per_kb,
            size=t.size,
            txid=t.txid,
            txhash=t.txhash,
            date=t.date,
            confirmations=t.confirmations,
            block_height=t.block_height,
            block_hash=t.block_hash,
            input_total=t.input_total,
            output_total=t.output_total,
            rawtx=t.rawtx,
            status=t.status,
            coinbase=t.coinbase,
            verified=t.verified,
            flag=t.flag,
            expiry_height=t.expiry_height,
        )

    @classmethod
    async def from_txid(cls, hdwallet: Wallet, txid: str):
        """
        Read single transaction from database with given transaction ID / transaction hash

        :param hdwallet: Wallet object
        :type hdwallet: Wallet
        :param txid: Transaction hash as hexadecimal string
        :type txid: str, bytes

        :return WalletClass:

        """
        async with hdwallet.db.get_session() as session:
            # If txid is unknown add it to database, else update
            res = await session.scalars(
                select(DbTransaction).filter(
                    DbTransaction.wallet_id == hdwallet.wallet_id,
                    DbTransaction.txid == to_bytes(txid),
                )
            )
            db_tx = res.first()

            if not db_tx:
                return

            await db_tx.awaitable_attrs.inputs
            await db_tx.awaitable_attrs.outputs

            fee_per_kb = None
            if db_tx.fee and db_tx.size:
                fee_per_kb = int((db_tx.fee / db_tx.size) * 1000)

            network = Network(db_tx.network_name)

            inputs = []
            for inp in db_tx.inputs:
                sequence = 0xFFFFFFFF
                if inp.sequence:
                    sequence = inp.sequence
                inp_keys = []

                if inp.key_id:
                    key = await hdwallet.key(inp.key_id)

                    if key.key_type == "multisig":
                        res = await session.scalars(
                            select(DbKey).filter_by(id=key.key_id)
                        )
                        db_key = res.first()

                        for ck in db_key.multisig_children:
                            inp_keys.append(ck.child_key.public.hex())
                    else:
                        inp_keys = key.key()

            inputs.append(
                Input(
                    prev_txid=inp.prev_txid,
                    output_n=inp.output_n,
                    keys=inp_keys,
                    unlocking_script=inp.script,
                    script_type=inp.script_type,
                    sequence=sequence,
                    index_n=inp.index_n,
                    value=inp.value,
                    double_spend=inp.double_spend,
                    witness_type=inp.witness_type,
                    network=network,
                    address=inp.address,
                    witnesses=inp.witnesses,
                )
            )

        outputs = []
        for out in db_tx.outputs:
            address = ""
            public_key = b""

            if out.key_id:
                key = await hdwallet.key(out.key_id)
                address = key.address

                if key.key_type != "multisig":
                    if key.key() and not isinstance(key.key(), Address):
                        public_key = key.key().public_hex

            outputs.append(
                Output(
                    value=out.value,
                    address=address,
                    public_key=public_key,
                    lock_script=out.script,
                    spent=out.spent,
                    output_n=out.output_n,
                    script_type=out.script_type,
                    network=network,
                )
            )

        return await cls.create(
            hdwallet=hdwallet,
            inputs=inputs,
            outputs=outputs,
            locktime=db_tx.locktime,
            version=db_tx.version,
            network=network,
            fee=db_tx.fee,
            fee_per_kb=fee_per_kb,
            size=db_tx.size,
            txid=to_hexstring(txid),
            date=db_tx.date,
            confirmations=db_tx.confirmations,
            block_height=db_tx.block_height,
            input_total=db_tx.input_total,
            output_total=db_tx.output_total,
            rawtx=db_tx.raw,
            status=db_tx.status,
            coinbase=db_tx.coinbase,
            verified=db_tx.verified,
        )

    def to_transaction(self):
        return Transaction(
            self.inputs,
            self.outputs,
            self.locktime,
            self.version,
            self.network.name,
            self.fee,
            self.fee_per_kb,
            self.size,
            self.txid,
            self.txhash,
            self.date,
            self.confirmations,
            self.block_height,
            self.block_hash,
            self.input_total,
            self.output_total,
            self.rawtx,
            self.status,
            self.coinbase,
            self.verified,
            self.witness_type,
            self.flag,
        )

    def sign(
        self,
        keys=None,
        index_n=0,
        multisig_key_n=None,
        hash_type=SIGHASH_ALL,
        fail_on_unknown_key=False,
        replace_signatures=False,
    ):
        """
        Sign this transaction. Use existing keys from wallet or use keys argument for extra keys.

        :param keys: Extra private keys to sign the transaction
        :type keys: HDKey, str
        :param index_n: Transaction index_n to sign
        :type index_n: int
        :param multisig_key_n: Index number of key for multisig input for segwit transactions. Leave empty if not known. If not specified all possibilities will be checked
        :type multisig_key_n: int
        :param hash_type: Hashtype to use, default is SIGHASH_ALL
        :type hash_type: int
        :param fail_on_unknown_key: Method fails if public key from signature is not found in public key list
        :type fail_on_unknown_key: bool
        :param replace_signatures: Replace signature with new one if already signed.
        :type replace_signatures: bool

        :return None:
        """
        priv_key_list_arg = []
        if keys:
            key_paths = list(
                dict.fromkeys(
                    [ti.key_path for ti in self.inputs if ti.key_path[0] == "m"]
                )
            )
            if not isinstance(keys, list):
                keys = [keys]

            for priv_key in keys:
                if not isinstance(priv_key, HDKey):
                    if isinstance(priv_key, str) and len(str(priv_key).split(" ")) > 4:
                        priv_key = HDKey.from_passphrase(priv_key, network=self.network)
                    else:
                        priv_key = HDKey(priv_key, network=self.network.name)

                priv_key_list_arg.append((None, priv_key))

                if key_paths and priv_key.depth == 0 and priv_key.key_type != "single":
                    for key_path in key_paths:
                        priv_key_list_arg.append(
                            (key_path, priv_key.subkey_for_path(key_path))
                        )

        for ti in self.inputs:
            priv_key_list = []
            for key_path, priv_key in priv_key_list_arg:
                if (
                    not key_path or key_path == ti.key_path
                ) and priv_key not in priv_key_list:
                    priv_key_list.append(priv_key)
            priv_key_list += [k for k in ti.keys if k.is_private]
            Transaction.sign(
                self,
                priv_key_list,
                ti.index_n,
                multisig_key_n,
                hash_type,
                fail_on_unknown_key,
                replace_signatures,
            )
        self.verify()
        self.error = ""

    async def send(self, offline: bool = False) -> None:
        """
        Verify and push transaction to network. Update UTXO's in database after successful send

        :param offline: Just return the transaction object and do not send it when offline = True. Default is False
        :type offline: bool

        :return None:

        """

        self.error = None
        if not self.verified and not self.verify():
            self.error = "Cannot verify transaction"
            return None

        if offline:
            return None

        srv = Service(
            network=self.hdwallet.network.name,
            providers=self.hdwallet.providers,
            cache_uri=self.hdwallet.db_cache_uri,
        )
        res = await srv.sendrawtransaction(self.raw_hex())

        if not res:
            self.error = "Cannot send transaction. %s" % srv.errors
            return

        if "txid" in res:
            _logger.info("Successfully pushed transaction, result: %s" % res)
            self.txid = res["txid"]
            self.status = "unconfirmed"
            self.confirmations = 0
            self.pushed = True
            self.response_dict = srv.results
            await self.store()

            async with self.hdwallet.db.get_session() as session:
                # Update db: Update spent UTXO's, add transaction to database
                for inp in self.inputs:
                    txid = inp.prev_txid
                    res = await session.scalars(
                        select(DbTransactionOutput)
                        .join(DbTransaction)
                        .filter(
                            DbTransaction.txid == txid,
                            DbTransactionOutput.output_n == inp.output_n_int,
                            DbTransactionOutput.spent.is_(False),
                        )
                    )
                    utxos = res.all()

                    for u in utxos:
                        u.spent = True

                await session.commit()

            await self.hdwallet._balance_update(network=self.network.name)
            return

        self.error = "Transaction not sent, unknown response from service providers"

    async def store(self):
        """
        Store this transaction to database

        :return int: Transaction index number
        """
        async with self.hdwallet.db.get_session() as session:
            # If txid is unknown add it to database, else update
            res = await session.scalars(
                select(DbTransaction).filter(
                    DbTransaction.wallet_id == self.hdwallet.wallet_id,
                    DbTransaction.txid == bytes.fromhex(self.txid),
                )
            )
            db_tx = res.first()

            if not db_tx:
                res = await session.scalars(
                    select(DbTransaction).filter(
                        DbTransaction.wallet_id.is_(None),
                        DbTransaction.txid == bytes.fromhex(self.txid),
                    )
                )
                db_tx = res.first()

                if db_tx:
                    db_tx.wallet_id = self.hdwallet.wallet_id

            rolled_back = False
            if not db_tx:
                version = 4 if self.network.name == "flux" else 1
                new_tx = DbTransaction(
                    wallet_id=self.hdwallet.wallet_id,
                    version=version,
                    txid=bytes.fromhex(self.txid),
                    block_height=self.block_height,
                    size=self.size,
                    confirmations=self.confirmations,
                    date=self.date,
                    fee=self.fee,
                    status=self.status,
                    input_total=self.input_total,
                    output_total=self.output_total,
                    network_name=self.network.name,
                    raw=self.rawtx,
                    verified=self.verified,
                    account_id=self.account_id,
                    expiry_height=self.expiry_height,
                )
                session.add(new_tx)
                try:
                    await session.commit()

                except IntegrityError:
                    _logger.info(
                        f"Rolling back this transaction, already stored for tx: {self.txid}"
                    )
                    await session.rollback()
                    rolled_back = True

                except Exception as e:  # pragma: no cover
                    _logger.warning(f"Tx store failure tx: {e}")

                txidn = new_tx.id

            if db_tx or rolled_back:
                if rolled_back:
                    res = await session.scalars(
                        select(DbTransaction).filter(
                            DbTransaction.wallet_id == self.hdwallet.wallet_id,
                            DbTransaction.txid == bytes.fromhex(self.txid),
                        )
                    )
                    db_tx = res.first()
                    _logger.info(f"Found previously stored transaction: {self.txid}")

                txidn = db_tx.id
                db_tx.block_height = (
                    self.block_height if self.block_height else db_tx.block_height
                )
                db_tx.confirmations = (
                    self.confirmations if self.confirmations else db_tx.confirmations
                )
                db_tx.date = self.date if self.date else db_tx.date
                db_tx.fee = self.fee if self.fee else db_tx.fee
                db_tx.status = self.status if self.status else db_tx.status
                db_tx.input_total = (
                    self.input_total if self.input_total else db_tx.input_total
                )
                db_tx.output_total = (
                    self.output_total if self.output_total else db_tx.output_total
                )
                db_tx.network_name = (
                    self.network.name if self.network.name else db_tx.name
                )
                db_tx.raw = self.rawtx if self.rawtx else db_tx.raw
                db_tx.verified = self.verified
                # session.add(db_tx)
                await session.commit()

            # why
            assert txidn
            for ti in self.inputs:
                res = await session.scalars(
                    select(DbKey).filter_by(
                        wallet_id=self.hdwallet.wallet_id, address=ti.address
                    )
                )
                tx_key = res.first()

                key_id = None
                if tx_key:
                    key_id = tx_key.id
                    tx_key.used = True

                res = await session.scalars(
                    select(DbTransactionInput).filter_by(
                        transaction_id=txidn, index_n=ti.index_n
                    )
                )
                tx_input = res.first()

                if not tx_input:
                    witnesses = int_to_varbyteint(len(ti.witnesses)) + b"".join(
                        [bytes(varstr(w)) for w in ti.witnesses]
                    )
                    new_tx_item = DbTransactionInput(
                        transaction_id=txidn,
                        output_n=ti.output_n_int,
                        key_id=key_id,
                        value=ti.value,
                        prev_txid=ti.prev_txid,
                        index_n=ti.index_n,
                        double_spend=ti.double_spend,
                        script=ti.unlocking_script,
                        script_type=ti.script_type,
                        witness_type=ti.witness_type,
                        sequence=ti.sequence,
                        address=ti.address,
                        witnesses=witnesses,
                    )
                    session.add(new_tx_item)
                elif key_id:
                    tx_input.key_id = key_id
                    if ti.value:
                        tx_input.value = ti.value
                    if ti.prev_txid:
                        tx_input.prev_txid = ti.prev_txid
                    if ti.unlocking_script:
                        tx_input.script = ti.unlocking_script

                await session.commit()
            for to in self.outputs:
                res = await session.scalars(
                    select(DbKey).filter_by(
                        wallet_id=self.hdwallet.wallet_id, address=to.address
                    )
                )
                tx_key = res.first()
                key_id = None

                if tx_key:
                    key_id = tx_key.id
                    tx_key.used = True

                spent = to.spent
                res = await session.scalars(
                    select(DbTransactionOutput).filter_by(
                        transaction_id=txidn, output_n=to.output_n
                    )
                )
                tx_output = res.first()

                if not tx_output:
                    new_tx_item = DbTransactionOutput(
                        transaction_id=txidn,
                        output_n=to.output_n,
                        key_id=key_id,
                        address=to.address,
                        value=to.value,
                        spent=spent,
                        script=to.lock_script,
                        script_type=to.script_type,
                    )
                    session.add(new_tx_item)
                elif key_id:
                    tx_output.key_id = key_id
                    tx_output.spent = spent if spent is not None else tx_output.spent
                await session.commit()
        return txidn

    def info(self):
        """
        Print Wallet transaction information to standard output. Include send information.
        """

        BaseTransaction.info(self)
        print("Pushed to network: %s" % self.pushed)
        print("Wallet: %s" % self.hdwallet.name)
        if self.error:
            print("Errors: %s" % self.error)
        print("\n")

    def export(self, skip_change=True):
        """
        Export this transaction as list of tuples in the following format:
            (transaction_date, transaction_hash, in/out, addresses_in, addresses_out, value, fee)

        A transaction with multiple inputs or outputs results in multiple tuples.

        :param skip_change: Do not include outputs to own wallet (default). Please note: So if this is set to True, then an internal transfer is not exported.
        :type skip_change: boolean

        :return list of tuple:
        """
        mut_list = []
        wlt_addresslist = self.hdwallet.addresslist()
        input_addresslist = [i.address for i in self.inputs]
        if self.outgoing_tx:
            fee_per_output = self.fee / len(self.outputs)
            for o in self.outputs:
                o_value = -o.value
                if o.address in wlt_addresslist:
                    if skip_change:
                        continue
                    elif self.incoming_tx:
                        o_value = 0
                mut_list.append(
                    (
                        self.date,
                        self.txid,
                        "out",
                        input_addresslist,
                        o.address,
                        o_value,
                        fee_per_output,
                    )
                )
        else:
            for o in self.outputs:
                if o.address not in wlt_addresslist:
                    continue
                mut_list.append(
                    (
                        self.date,
                        self.txid,
                        "in",
                        input_addresslist,
                        o.address,
                        o.value,
                        0,
                    )
                )
        return mut_list

    def save(self, filename=None):
        """
        Store transaction object as file, so it can be imported in fluxwallet later with the :func:`load` method.

        :param filename: Location and name of file, leave empty to store transaction in fluxwallet data directory: .fluxwallet/<transaction_id.tx)
        :type filename: str

        :return:
        """
        if not filename:
            p = Path(FW_DATA_DIR, "%s.tx" % self.txid)
        else:
            p = Path(filename)
            if not p.parent or str(p.parent) == ".":
                p = Path(FW_DATA_DIR, filename)
        f = p.open("wb")
        t = self.to_transaction()
        pickle.dump(t, f)
        f.close()

    def delete(self):
        """
        Delete this transaction from database.

        WARNING: Results in incomplete wallets, transactions will NOT be automatically downloaded again when scanning or updating wallet. In normal situations only used to remove old unconfirmed transactions

        :return int: Number of deleted transactions
        """

        session = self.hdwallet._session
        txid = bytes.fromhex(self.txid)
        tx_query = session.query(DbTransaction).filter_by(txid=txid)
        tx = tx_query.scalar()
        session.query(DbTransactionOutput).filter_by(transaction_id=tx.id).delete()
        session.query(DbTransactionInput).filter_by(transaction_id=tx.id).delete()
        session.query(DbKey).filter_by(latest_txid=txid).update(
            {DbKey.latest_txid: None}
        )
        res = tx_query.delete()
        self.hdwallet._commit()
        return res

    async def store(self):
        """
        Store this transaction to database

        :return int: Transaction index number
        """
        async with self.hdwallet.db.get_session() as session:
            # If txid is unknown add it to database, else update
            res = await session.scalars(
                select(DbTransaction).filter(
                    DbTransaction.wallet_id == self.hdwallet.wallet_id,
                    DbTransaction.txid == bytes.fromhex(self.tx.txid),
                )
            )
            db_tx = res.first()

            if not db_tx:
                res = await session.scalars(
                    select(DbTransaction).filter(
                        DbTransaction.wallet_id.is_(None),
                        DbTransaction.txid == bytes.fromhex(self.tx.txid),
                    )
                )
                db_tx = res.first()

                if db_tx:
                    db_tx.wallet_id = self.hdwallet.wallet_id

            rolled_back = False
            if not db_tx:
                # seems dodgey
                version = 4 if self.tx.network.name == "flux" else 1

                new_tx = DbTransaction(
                    wallet_id=self.hdwallet.wallet_id,
                    version=version,
                    txid=bytes.fromhex(self.tx.txid),
                    block_height=self.tx.block_height,
                    size=self.tx.size,
                    confirmations=self.tx.confirmations,
                    date=self.tx.date,
                    fee=self.tx.fee,
                    status=self.status,
                    input_total=self.tx.input_total,
                    output_total=self.tx.output_total,
                    network_name=self.tx.network.name,
                    raw=self.tx.rawtx,
                    verified=self.tx.verified,
                    account_id=self.account_id,
                    coinbase=self.tx.coinbase,
                    expiry_height=self.tx.expiry_height,
                )
                session.add(new_tx)
                try:
                    await session.commit()

                except IntegrityError:
                    print("ROLLEDBACK")
                    _logger.info(
                        f"Rolling back this transaction, already stored for tx: {self.tx.txid}"
                    )
                    await session.rollback()
                    rolled_back = True

                except Exception as e:  # pragma: no cover
                    _logger.warning(f"Tx store failure tx: {e}")

                txidn = new_tx.id

            if db_tx or rolled_back:
                if rolled_back:
                    res = await session.scalars(
                        select(DbTransaction).filter(
                            DbTransaction.wallet_id == self.hdwallet.wallet_id,
                            DbTransaction.txid == bytes.fromhex(self.tx.txid),
                        )
                    )
                    db_tx = res.first()
                    _logger.info(f"Found previously stored transaction: {self.tx.txid}")

                txidn = db_tx.id
                db_tx.block_height = (
                    self.tx.block_height if self.tx.block_height else db_tx.block_height
                )
                db_tx.confirmations = (
                    self.tx.confirmations
                    if self.tx.confirmations
                    else db_tx.confirmations
                )
                db_tx.date = self.tx.date if self.tx.date else db_tx.date
                db_tx.fee = self.tx.fee if self.tx.fee else db_tx.fee
                db_tx.status = self.status if self.status else db_tx.status
                db_tx.input_total = (
                    self.tx.input_total if self.tx.input_total else db_tx.input_total
                )
                db_tx.output_total = (
                    self.tx.output_total if self.tx.output_total else db_tx.output_total
                )
                db_tx.network_name = (
                    self.tx.network.name if self.tx.network.name else db_tx.network_name
                )
                db_tx.raw = self.tx.rawtx if self.tx.rawtx else db_tx.raw
                db_tx.verified = self.tx.verified
                db_tx.coinbase = self.tx.coinbase

                await session.commit()

            for ti in self.tx.inputs:
                res = await session.scalars(
                    select(DbKey).filter_by(
                        wallet_id=self.hdwallet.wallet_id, address=ti.address
                    )
                )
                tx_key = res.first()

                key_id = None
                if tx_key:
                    key_id = tx_key.id
                    tx_key.used = True

                res = await session.scalars(
                    select(DbTransactionInput).filter_by(
                        transaction_id=txidn, index_n=ti.index_n
                    )
                )
                tx_input = res.first()

                if not tx_input:
                    witnesses = int_to_varbyteint(len(ti.witnesses)) + b"".join(
                        [bytes(varstr(w)) for w in ti.witnesses]
                    )
                    new_tx_item = DbTransactionInput(
                        transaction_id=txidn,
                        output_n=ti.output_n_int,
                        key_id=key_id,
                        value=ti.value,
                        prev_txid=ti.prev_txid,
                        index_n=ti.index_n,
                        double_spend=ti.double_spend,
                        script=ti.unlocking_script,
                        script_type=ti.script_type,
                        witness_type=ti.witness_type,
                        sequence=ti.sequence,
                        address=ti.address,
                        witnesses=witnesses,
                    )
                    session.add(new_tx_item)
                elif key_id:
                    tx_input.key_id = key_id
                    if ti.value:
                        tx_input.value = ti.value
                    if ti.prev_txid:
                        tx_input.prev_txid = ti.prev_txid
                    if ti.unlocking_script:
                        tx_input.script = ti.unlocking_script

                await session.commit()

            for to in self.tx.outputs:
                res = await session.scalars(
                    select(DbKey).filter_by(
                        wallet_id=self.hdwallet.wallet_id, address=to.address
                    )
                )
                tx_key = res.first()
                key_id = None

                if tx_key:
                    key_id = tx_key.id
                    tx_key.used = True

                spent = to.spent
                res = await session.scalars(
                    select(DbTransactionOutput).filter_by(
                        transaction_id=txidn, output_n=to.output_n
                    )
                )
                tx_output = res.first()

                if not tx_output:
                    new_tx_item = DbTransactionOutput(
                        transaction_id=txidn,
                        output_n=to.output_n,
                        key_id=key_id,
                        address=to.address,
                        value=to.value,
                        spent=spent,
                        script=to.lock_script,
                        script_type=to.script_type,
                    )
                    session.add(new_tx_item)
                elif key_id:
                    tx_output.key_id = key_id
                    tx_output.spent = spent if spent is not None else tx_output.spent

                rolled_back = False
                try:
                    await session.commit()

                except IntegrityError:
                    print("ROLLEDBACK")
                    _logger.info(
                        f"Rolling back this transaction, already stored for tx: {self.tx.txid}"
                    )
                    await session.rollback()
                    rolled_back = True

                except Exception as e:  # pragma: no cover
                    _logger.warning(f"Tx store failure tx: {e}")

                if rolled_back and key_id:
                    res = await session.scalars(
                        select(DbTransactionOutput).filter_by(
                            transaction_id=txidn, output_n=to.output_n
                        )
                    )
                    tx_output = res.first()

                    tx_output.key_id = key_id
                    tx_output.spent = spent if spent is not None else tx_output.spent
                    await session.commit()

        return txidn
